import utils.displayStandards as dsp
from . import scattering_factors as scatf
import numpy as np
from math import pi

__all__=['structure_factor3D','plot_structure3D','get_miller3D',
        'get_pendulossung']

#2D for factors
ai = 1/np.array([0.1,0.25,0.26,0.27,1.5])**2
fj = lambda q,j,eps:eps*(np.pi/ai[j])*np.exp(-(np.pi*q)**2/ai[j])


def structure_factor3D(pattern,lat_vec,hkl=None,hklMax=10,sym=1,v=''):
    '''Computes structure factor in 3D from :
    - `pattern` : Nx4 array - N atoms with each row : fractional coordinates and Za
    - `lat_vec` : 3x3 array - reciprocal lattice vectors (2*pi/a convention)
    - `hkl`     : list of 3 miller indices h,k,l each as 3Ndarray
    - `hklMax`  : int - max miller index in each direction from -hklMax to hklMax
    '''
    #unpack
    if not hkl : hkl = get_miller3D(hklMax,sym)
    hx,ky,lz = hkl
    # print('rearranging h,k,l so h[(hi,k,l)]=hi')
    hx,ky,lz = np.transpose(hx,[1,0,2]),np.transpose(ky,[1,0,2]),np.transpose(lz,[1,0,2])
    hkl = [hx,ky,lz]
    ra,fa = pattern[:,:3],pattern[:,3]
    atoms = list(np.array(np.unique(fa),dtype=int));#print(atoms)
    # get scattering factor
    b1,b2,b3 = lat_vec
    k_x,k_y,k_z = hx*b1[0]+ky*b2[0]+lz*b3[0],hx*b1[1]+ky*b2[1]+lz*b3[1],hx*b1[2]+ky*b2[2]+lz*b3[2]
    q = np.sqrt(k_x**2+k_y**2+k_z**2)/(2*pi)
    q,fq = scatf.get_elec_atomic_factors(atoms,q)
    if 'q' in v:qmax=q.max();print('qmax=%.4f A^-1\nmax_res=%.4f A' %(qmax,1/qmax))
    #structure factor
    Fhkl,n_atoms = np.zeros(hx.shape,dtype=complex),len(atoms)
    for i,atom in zip(range(n_atoms),atoms):
        F_i = np.zeros(hx.shape,dtype=complex)
        idx = fa==atom
        for ri in ra[idx,:]:
            # print(ri)
            # print(ri,np.exp(-2*pi*1J*(ri[0]*hx+ri[1]*ky+ri[2]*lz)))
            F_i += np.exp(-2*pi*1J*(ri[0]*hx+ri[1]*ky+ri[2]*lz))
        Fhkl += F_i*fq[i]

    # cs=dsp.getCs('Spectral',n_atoms)
    # plts=[[q.flatten(),fq[i].flatten(),[cs[i],'+'],atom] for i,atom in enumerate(atoms)]
    # dsp.stddisp(plts,lw=2)
    return hkl,Fhkl

def structure_factor2D(pattern,lat_vec,hk=None,hkMax=10,sym=1,v=0,eps=1):
    '''get structure factor
    - pattern : atomic positions in real space (fractional coordinates)
    - lat_vec : reciprocal lattice vectors (2pi/a convention)
    '''
    #unpack
    if not hk : hl = get_miller2D(hkMax,sym)
    hx,ky = hl
    hx,ky = hx.T,ky.T
    hl = [hx,ky]
    ra,fa = pattern[:,:2],pattern[:,2]
    atoms = list(np.array(np.unique(fa),dtype=int));#print(atoms)
    #scattering factor
    b1,b2 = lat_vec
    k_x,k_y = hx*b1[0]+ky*b2[0],hx*b1[1]+ky*b2[1]
    q = np.sqrt(k_x**2+k_y**2)/(2*pi)
    fq = [fj(q,Za,eps) for Za in atoms]
    if v : qmax=q.max();print('qmax=%.4f A^-1\nmax_res=%.4f A' %(qmax,1/qmax))
    #compute structure factor
    Fhl,n_atoms = np.zeros(hx.shape,dtype=complex),len(atoms)
    for i,atom in zip(range(n_atoms),atoms):
        F_i = np.zeros(hx.shape,dtype=complex)
        idx = fa==atom
        for ri in ra[idx,:]:
            F_i += np.exp(2*pi*1J*(ri[0]*hx+ri[1]*ky))
        Fhl += F_i*fq[i]
    return hl,Fhl

###
def get_pendulossung(name='Si',miller=[0,0,0],keV=200,opt='p'):
    ''' give the theorertical 2-beam approximation Pendullosung thickness
    - `name`    : compound
    - `miller`  : [h,k,l]
    - `keV`     : wavelength (keV)
    RETURN
    - xi : Pendullosung thickness (A)
    '''
    # compute structure factor
    from crystals import Crystal
    crys = Crystal.from_database(name)
    lat_vec  = crys.reciprocal_vectors
    pattern  = np.array([list(a.coords_fractional)+[a.atomic_number] for a in crys.atoms])
    hkl,Fhkl = structure_factor3D(pattern,lat_vec,hklMax=max(miller),sym=0,v='')
    ax,by,cz = crys.lattice_parameters[:3]
    Vcell    = crys.volume
    # compute Pendullosung
    h,k,l = miller
    Ug,K  = np.abs(Fhkl[h,k,l])/Vcell,1/scatf.wavelength(keV)
    xi  = K/Ug
    if 'p' in opt:
        print(green+"\tPendullosung thickness "+name+'[%d%d%d]' %(h,k,l)+black)
        print('%-5s= %.2f\n%-5s= %.2f\n%-5s= %.2f'%('K',K,'Ug',Ug,'xi',xi))
    return xi

##########################################################################
###def : display
##########################################################################
def plot_structure3D(hkl,Fhkl,log=0,**kwargs):
    '''scatter plot of the intensity'''
    hx,ky,lz = hkl
    hx,ky,lz = hx.flatten(),ky.flatten(),lz.flatten()
    S        = np.abs(Fhkl.flatten())**2
    if log : S = np.log10(S+1)
    N = hx.max()
    lims = [0,N,0,N,0,N]
    dsp.stddisp(scat=[hx,ky,lz,S],rc='3d',xylims=lims,**kwargs)

def plot2Dcutplane(Shkl,n='100',title='auto',**kwargs):
    '''Not working for arbitrary cut yet '''
    N = Shkl.shape[0]
    if title=='auto' : title='Silicon([%s]) $S_{hk}$' %n
    if   n=='100' : Scut = Shkl[0,:,:]
    elif n=='110' : Scut = Shkl[np.identity(N,dtype=bool),:]
    dsp.stddisp(im=Scut,imOpt='c',legOpt=0,title=title,**kwargs)

def plot1Dcutline(Shkl,u='001',lat_vec=None,dopt='fq',**kwargs):
    '''
    u : str cutline
    dopt:f(fe(q)),q(q units)
    '''
    N,labx = Shkl.shape[0],'$index$'
    ql = np.arange(N)
    if u == '001' :
        Scut = Shkl[0,0,:]
    else :
        Scut = Shkl[0,0,:]
    plts = [[ql,Scut,'bs-','$S_{%s}$' %u]]
    if 'q' in dopt or 'f' in dopt :
        b1,b2,b3 = lat_vec
        ql *= b3[2]/(2*pi)
        labx = '$q(A^{-1})$'
    if 'f' in dopt:
        #k_x,k_y,k_z = hx*b1[0]+ky*b2[0]+lz*b3[0],hx*b1[1]+ky*b2[1]+lz*b3[1],hx*b1[2]+ky*b2[2]+lz*b3[2]
        #q = np.sqrt(k_x**2+k_y**2+k_z**2)/(2*pi)
        qz = np.linspace(0,N,1000)*b3[2]/(2*pi)
        qz,fq = scatf.get_elec_atomic_factors([14],qz)
        plts += [[qz,64*fq[0]**2,'g--','$64f_e^2$'],[qz,32*fq[0]**2,'c--','$32f_e^2$']]
    dsp.stddisp(plts,labs=[labx,'$S_q(A^2)$'],**kwargs)

##########################################################################
##### def: misc
def get_miller3D(hklMax=10,sym=1):
    '''hklMax : Number of Beams such that Fourier components = 2^(N+1) - 1'''
    Nhkl = range(-hklMax*sym,hklMax+1)
    hkl = np.meshgrid(Nhkl,Nhkl,Nhkl)
    return hkl
def get_miller2D(hkMax=10,sym=1):
    Nhk = range(-hkMax*sym,hkMax+1)
    hk = np.meshgrid(Nhk,Nhk)
    return hk

##########################################################################
#### def : test
##########################################################################
def _test_structure_factor2D(N=5):
    pptype,a,b,angle='p1',50,50,90      #A
    pattern = np.array([[0.5,0.5,1]])

    p1=pg.Wallpaper(pptype,a,b,angle,pattern=pattern,gen=False,pOpt='')
    b1,b2 = p1.get_reciprocal_lattice_2D()

    hl = np.arange(-2**N+1,2**N)
    h,l = np.meshgrid(hl,hl)
    Fhl = get_structure_factor2D(pattern,h,l,[b1,b2])

    qx,qy = (h*b1[0]+l*b2[0])/(2*pi),(h*b1[1]+l*b2[1])/(2*pi)
    fig,(axM,axP) = create_fig('12',pad=0.5,rc='12')
    axM.pcolor(np.abs(Fhl)**2);
    axP.pcolor(qx,qy,np.abs(Fhl)**2);
    #axP.pcolor(kx,ky,np.angle(Fhl));
    fig.show()
    return h,l,qx,qy,Fhl

def _test_structure_factor3D(N=5,sym=False,**kwargs):
    from crystals import Crystal
    Si = Crystal.from_database('Si'); #vol     = Si.volume
    lat_vec = Si.reciprocal_vectors
    rj      = np.array([a.coords_fractional for a in Si.atoms])
    pattern = np.concatenate((rj,14*np.ones((rj.shape[0],1))),axis=1)
    hkl,Fhkl  = structure_factor3D(pattern,lat_vec,hklMax=N,sym=sym)
    Shkl = np.abs(Fhkl)**2
    #plot_structure3D(hkl,Fhkl,name=figpath+'Si_Shkl.png',opt='s',cmap='jet',ms=50,title='Si $c=log_{10}(S+1)$',view=[9,-72])
    Shkl = np.log10(Shkl+1)
    plot2Dcutplane(Shkl,n='100',name=figpath+'Si_S100.png',**kwargs)
    plot2Dcutplane(Shkl,n='110',name=figpath+'Si_S110.png',**kwargs)
    #plot1Dcutline(Shkl,u='001',dopt='')
    return hkl,Fhkl


if __name__ == "__main__" :
    figpath=get_figpath(__file__,"/figures/")
    plt.close('all')
    #hk,Fhk = _test_structure_factor2D(N=8)
    #hkl,Fhkl = _test_structure_factor3D(N=4,opt='p',cmap='jet')



import importlib as imp
import os,glob,io,re, numpy as np, pandas as pd, tifffile,mrcfile,scipy.optimize as opt
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from subprocess import Popen,PIPE,check_output
from crystals import Crystal,Lattice
from utils import physicsConstants as cst
from . import utilities as ut;imp.reload(ut)
from . import import_ED as ED               ;imp.reload(ED)

class XDS(ED.Dataset):
    def __init__(self,xds_ascii:str,):
        self.read_xds_ascii(xds_ascii)

        # A is the orientation matrix in lab space (not reciprocal space)
        self.A = np.array([self.info[s] for s in
                ['UNIT_CELL_A-AXIS','UNIT_CELL_B-AXIS','UNIT_CELL_C-AXIS']
            ])
        self.lam = self.info['X-RAY_WAVELENGTH']
        self.F   = self.info['DETECTOR_DISTANCE']
        self.orgx,self.orgy = np.array([self.info['ORGX'],self.info['ORGY']])
        self.dx,self.dy = self.info['QX'],self.info['QY']
        self.nxy = np.array([self.info['NX'],self.info['NY']],dtype=int)
        self.lat_params = self.info["UNIT_CELL_CONSTANTS"]

        self.init_geom()
        self.get_uvw()

        # lat = Lattice.from_parameters(*self.info["UNIT_CELL_CONSTANTS"])
        # Alat=np.array(lat.lattice_vectors).T
        # Arec=np.array(lat.reciprocal_vectors).T/(2*np.pi)
        # self.Arec=Arec
        # self.R0 = np.linalg.inv(Alat).dot(self.A.T)
        # self.UB=Arec.dot(np.linalg.inv(Alat).dot(self.A.T))
        #
        # #the proper way
        # a,b,c = self.A
        # a_star = np.cross(b,c)/a.dot(np.cross(b,c))
        # b_star = np.cross(c,a)/b.dot(np.cross(c,a))
        # c_star = np.cross(a,b)/c.dot(np.cross(a,b))
        # self.UB2 = np.array([a_star,b_star,c_star]).T

        # print(Alat)
        # print(self.A)
        # Alab=np.linalg.inv(Alat).dot(self.A)
        # print(Alab,np.linalg.norm(Alab,axis=1))
        # print(Arec)
        # print(self.UB)


        self.get_qxy()

    def xd_to_px(self,xd):
        px = xd[:,0]/self.dx + self.orgx            #;print('px:',px)
        py = xd[:,1]/self.dy + self.orgy            #;print('py:',py)
        return px,py

    def read_xds_ascii(self,xds_ascii):
        self.path=os.path.dirname(xds_ascii)
        with open(xds_ascii,'r') as f:
            lines = f.readlines()
        info = dict()
        keys = iter(['DATA_RANGE','ROTATION_AXIS','OSCILLATION_RANGE','STARTING_ANGLE',
            'UNIT_CELL_CONSTANTS','UNIT_CELL_A-AXIS','UNIT_CELL_B-AXIS','UNIT_CELL_C-AXIS',
            'X-RAY_WAVELENGTH','INCIDENT_BEAM_DIRECTION','NX','ORGX','DETECTOR_DISTANCE',
            'DIRECTION_OF_DETECTOR_X-AXIS','DIRECTION_OF_DETECTOR_Y-AXIS',
            'END_OF_HEADER',
            ])
        lines=iter(lines)
        l,k=next(lines),next(keys)
        while '!END_OF_HEADER' not in l:
            # print(l,k)
            l = re.sub("\s{2,}"," ",l[1:].strip())
            if k in l:
                v=l.split("=")
                if len(v)>2:
                    v = ''.join(v).split(" ")
                    names,vals = v[::2],v[1::2]
                    for key,val in zip(names,vals):
                        info[key] = float(val)
                else:
                    val=np.array(v[1].split(" ")[1:],dtype=float)
                    if val.size==1:val=val[0]
                    info[k]=val
                k=next(keys)
            l = next(lines)
        # # print(info)

        self.info=info
        self.info["DATA_RANGE"]=np.array(info["DATA_RANGE"],dtype=int)

        #detector
        e1,e2 = self.info['DIRECTION_OF_DETECTOR_X-AXIS'],self.info['DIRECTION_OF_DETECTOR_Y-AXIS']
        e3 = np.cross(e1,e2)
        ED = np.array([e1,e2,e3]).T
        self.ED = ED

        lines=''.join(list(lines)[:-1])
        hkl = pd.read_csv(io.StringIO(lines),engine="python",delimiter=' +',
            names = ['h','k','l','I','sigma','px','py','pz','rlp','peak','corr','psi'],
            )
        hkl.index = [str(tuple(h)) for h in hkl[['h','k','l']].values ]
        hkl['F'] = hkl.pz.round()
        hkl['hkl'] = hkl.index
        self.rpl = hkl[['h','k','l','I','px','py','pz','F','hkl']].copy()
        self.hkl = self.rpl[['I']].copy()

    def get_qxy(self):
        ED = self.ED
        ix,iy = self.rpl.px,self.rpl.py
        orgx,orgy = self.orgx,self.orgy
        qx,qy = self.dx,self.dy
        lam = self.lam
        x = qx*(ix-orgx)*ED[0,0] + qy*(iy-orgy)*ED[0,1] + self.F*ED[0,2]
        y = qx*(ix-orgx)*ED[1,0] + qy*(iy-orgy)*ED[1,1] + self.F*ED[1,2]
        z = qx*(ix-orgx)*ED[2,0] + qy*(iy-orgy)*ED[2,1] + self.F*ED[2,2]
        s = np.array([x,y,z])
        # print('s:\n',s[:,:5].T)
        s_norms=np.linalg.norm(s,axis=0)
        # print('s norms : \n',s_norms[:5])
        s = (s/s_norms).T
        # s/=lam
        # print('s normalized : \n',s[:5,:])
        self.rpl[['s1_x','s1_y','s1_z']] = s
        # s0 = np.array([0,0,1])
        s0 = self.info['INCIDENT_BEAM_DIRECTION']
        r  = (s-s0)/lam
        # print('r:\n',r[:5,:])
        self.rpl[['r_x','r_y','r_z']] = r
        self.rpl[['qx','qy']] = self.rpl[['r_x','r_y']]



import numpy as np
import pandas as pd
from utils import displayStandards as dsp
from utils import glob_colors as colors
from utils.physicsConstants import mc2,c0,A,h,emass,keV

dat_path = dsp.get_figpath(__file__,'/data/')
__all__ = ['get_xray_atomic_factors','get_elec_atomic_factors',
    'get_v_from_KE','wavelength','get_elt']


def get_xray_atomic_factors(elts,qmax=2,npts=100):
    '''elts : list of atoms by name
    The formula on the uses the semi scattering angle
    '''
    q = np.linspace(0,min(qmax,6),npts) #A
    #q2 = ((q/(4*pi))**2)[:,None]
    q2 = ((q/2)**2)[:,None]
    dfx = pd.read_pickle(dat_path+'xray.pkl').loc[elts].astype(float)
    Ai = dfx[['a%d' %i for i in range(1,5)]].values
    Bi = dfx[['b%d' %i for i in range(1,5)]].values
    C  = dfx['c'].values
    fx_q = [(ai*np.exp(-bi*q2)).sum(axis=1) + c  for ai,bi,c in zip(Ai,Bi,C)]
    return q,fx_q

def get_fe(Zs,q):
    ''' computes th atomic form factors
    - Z : atomic numbers
    - q : reciprocal vectors
    returns :
    - fq_e : nZ x Nqs
    '''
    fparams = np.load(dat_path+'abcd.npy',allow_pickle=True)
    q2 = q**2
    fq_e = np.zeros((q.size,Zs.size))
    for iZ,Z in enumerate(Zs):
        a,b=np.reshape(fparams[Z,:6],(3,2)).T#;
        c,d=np.reshape(fparams[Z,6:],(3,2)).T#;print(a,b,c,d)
        fq = np.zeros(q.shape)
        for i in range(3):
             fq += a[i]/(b[i]+q2)+c[i]*np.exp(-d[i]*q2)
        fq_e[:,iZ]=fq
    return fq_e


def get_elec_atomic_factors(elts,q=None,qmax=2,npts=100):
    '''
    elts : list of atoms by name 'E' or atomic number Z
    opt : G(gauss fit), L(Lenz)
    '''
    if isinstance(elts[0],str) :
        Z = pd.read_pickle(dat_path+'elec.pkl')['Z'][elts].values
    else:
        Z=elts
    if not isinstance(q,np.ndarray) : q = np.linspace(0,qmax,npts)
    #dummy case to disable form factor
    # if Z==[0]:return q,[np.ones(q.shape)]
    q2,fq_e,nelts = q**2,[],len(Z)

    fparams = np.load(dat_path+'abcd.npy',allow_pickle=True)
    for elt in range(nelts) :
        a,b=np.reshape(fparams[Z[elt],:6],(3,2)).T
        c,d=np.reshape(fparams[Z[elt],6:],(3,2)).T; #print(a,b,c,d)
        fq = np.zeros(q.shape)
        for i in range(3):
             fq += a[i]/(b[i]+q2)+c[i]*np.exp(-d[i]*q2)
        fq_e+=[fq]
    return q,fq_e

# def get_lenz_model(Z,q,E0):
#     T  = E0*(1+E0/(2*mc2))/(1+E0/(2*mc2))   #keV
#     w  = wavelength(E0)                     #A
#     k0 = 2*pi/w                             #A^-1
#     theta  = q*w                            #no dim
#     theta0 = Z**(1/3)/(k0*a0)               #no dim
#     return 2*Z/(sqrt(T)*k0)/(theta**2+theta0**2)

########################################################################
####def : misc
########################################################################
def get_v_from_KE(KE):
    ''' v = get_v_from_KE(KE)
    KE : energy (keV)
    v  : speed (normalized to c)
    '''
    v =  np.sqrt(1 - 1/(1+np.array(KE)/mc2)**2)
    return v

def wavelength(KE):
    '''w = wavelength(KE)
    KE : energy (keV)
    lam  : wavelength (Angstrom)
    '''
    #emass = 510.99906   #keV
    lam = h*c0/(np.sqrt(KE*(2*emass+KE))*keV)
    return lam/A

def get_elt(elts):
    Z = np.array(pd.read_pickle(dat_path+'elec.pkl')['Z'][elts].values,dtype=int)
    return Z

################################################################################
#### def : test
################################################################################
def _test_elect_atomic(name='',fmt='svg',**kwargs):
    elts = ['H','C','N','O','S','P']  #+ ['Si','Cu','Au','U']
    qe,fq_e=get_elec_atomic_factors(elts,qmax=3,npts=100)
    qx,fq_x=get_xray_atomic_factors(elts,qmax=6,npts=100)
    Z = get_elt(elts) #pd.read_pickle(dat_path+'atoms.pkl')['Z'][elts].values
    cs=[dsp.unicolor(0.85),dsp.unicolor(0.4),(0,0,1),(1,0,0),(0,1,1),(1,0,1)] #+ ['k']*4
    df=pd.DataFrame(np.array([Z,cs,fq_e,fq_x]).T,columns=['Z','color','fq_e','fq_x'],index=elts)


    plts=[[qe,elt.fq_e,elt.color,'$%s$' %elt.name ] for index,elt in df.iterrows()]
    dsp.stddisp(plts,labs=['$q(A^{-1})$','$f^e(A)$'],lw=2,
        name=name+'ED_scattering_factors.%s' %fmt,**kwargs)
    plts=[[qx,elt.fq_x,elt.color,'$%s$' %elt.name ] for index,elt in df.iterrows()]
    dsp.stddisp(plts,labs=['$q(A^{-1})$','$f^x(e)$'],lw=2,
        name=name+'MX_scattering_factors.%s' %fmt,**kwargs)
    return df

def _test_misc(KE=200,elts=['H','C','O','Si','Fe']):
    print( '''KE=%.2f,v=%.4f,lambda=%.4f
        ''' %(KE,get_v_from_KE(KE),wavelength(KE)))
    print(colors.green+'get_elt(',elts,') = ',colors.red,get_elt(elts), colors.black)
if __name__=="__main__":
    figpath='../multislice/docs_fig/'
    _test_elect_atomic(name=figpath,fmt='png',legLoc='upper right',opt='ps',fonts={'leg':35,'lab':40,'tick':30},figopt='t')
    _test_misc()



import pickle5,os,glob
import matplotlib.pyplot as plt
from matplotlib import rc
import pandas as pd
import numpy as np
from utils import displayStandards as dsp
from utils import glob_colors as colors
from utils.glob_colors import*
from scattering import scattering_factors as scat
# import get_elec_atomic_factors,wavelength
# from scattering.structure_factor import structure_factor3D

hosts = ['','brno','tarik-CCP4home']
cpus  = dict(zip(hosts,['asus-$i_5^{(4cores)}$','brno-$i_7^{(8cores)}$','xps-$i_7^{(8cores)}$']))
ncpus = dict(zip(hosts,[4,8,8]))
info_cols = ['zmax(A)','Inorm','cputime(s)','walltime(s)']

#########################################################################
### import from TEMSIM
#########################################################################
def import_beams(file,slice_thick=1,iBs=[],tol=1e-2,Obeam=False,pImax=False):
    ''' Import beam files with some postprocessing option\n
    - slice_thick : slice thickness used for simulation
    - iBs   : selected beam indices : default=Ibeams.max()>tol
    - tol   : select beams above threshold if iBs==[]
    - Obeam : include origin
    '''
    hk    = open(file).readline().rstrip().split('=  ')[1].split(' ');#print(hk)
    #print(hk)
    beams = np.loadtxt(file,skiprows=3).T
    if isinstance(slice_thick,list) :
        idx,beams = beams[0,:],beams[1:,:]
        n_slices = int(len(idx)/len(slice_thick)); #print(n_slices,len(slice_thick))
        t = np.cumsum(np.tile(slice_thick,n_slices))+0.75*slice_thick;#print(t[:4])
    else :
        idx,beams = beams[0,:-5],beams[1:,:-5]
        t = idx*slice_thick+0.75*slice_thick
    #get re,im,I
    nbs = len(hk)
    re = [beams[2*i,:] for i in range(nbs)]
    im = [beams[2*i+1,:] for i in range(nbs)]
    Ib = [np.abs(r+1J*i)**2 for r,i in zip(re,im)]
    #filter
    Imax = np.array([I.max() for I in Ib]); #print(Imax,Imax.max())
    if any(iBs) :
        if isinstance(iBs[0],str) :
            hk = np.array(hk)
            iBs = [ (iB==hk).argmax() for iB in iBs if (iB==hk).sum()]
    else :
        iBs = [i for i in range([1,0][Obeam],nbs) if Imax[i]>= tol*Imax.max()]
    if pImax : print('Imax:',Imax,'iBs:',iBs,', nbs=',nbs)
    hk,t,re,im,Ib = np.array(hk)[iBs],t, np.array(re)[iBs],np.array(im)[iBs],np.array(Ib)[iBs]
    return hk,t,re,im,Ib

#########################################################################
#### def : DataFrame utilities
#########################################################################
def rock_load(datpath,tag=''):
    if tag:tag+='_'
    with open(datpath+tag+'rock.pkl','rb') as f : rock = pickle5.load(f)
    return rock

def load(datpath,tail='',tag=None,v=0):
    if tag:tail=tag
    filename = datpath
    pkls = glob.glob(os.path.join(datpath,'*.pkl'))
    if any(pkls):
        filename = pkls[0]
        tails = ['_'.join(pkl.split('_')[1:-1]) for pkl in pkls]
        file_idx = [i for i,t in enumerate(tails) if tail == t]
        if file_idx:
            if len(file_idx)>1:print(colors.red+'several matches for tail : ',file_dx+colors.black)
            filename=pkls[file_idx[0]]
            print(colors.green+'loading ' +colors.yellow+filename+colors.black )
            multi = load_multi_obj(filename)
            if v : print('simu status : ',multi.check_simu_state())
            if v>1 : multi.log_info(v=1);
            return multi
        else:
            print(colors.red+'warning: tail "%s" not found in %s' %(tail,datpath)+colors.black )
            print(colors.green+'available tails:'+colors.yellow,tails,colors.black)
    else:
        print(colors.red+'warning:no file found'+colors.red)

def load_multi_obj(filename):
    '''load a saved Multislice object
    filename : pickle file (.pkl)  '''
    with open(filename,'rb') as f : multi = pickle5.load(f)
    return multi

def get_info(log_file):
    '''compute zmax,I,cpuTime and wallTime'''
    with open(log_file,'r') as f : log_lines=f.readlines()
    log_lines = log_lines[-2:]
    info = [l.strip().split('=')[-1].split(' ')[1] for l in log_lines]
    return np.array(info,dtype=float)

def update_df_info(df_path,hostpath=None,files=[]):
    df = pd.read_pickle(df_path)
    datpath = os.path.dirname(df_path)+'/'
    for name in df.index:
        multi = load_multi_obj(datpath+name);#print(name)
        ssh_host = df.loc[name].host
        state    = multi.check_simu_state(ssh_alias=ssh_host,v=0,hostpath=hostpath)
        df.loc[name]['state'] = state
        info = multi.log_info(v=0)
        df.loc[name][info_cols[:2]] = info[2:]
        if state=='done':
            info = multi.log_info(v=0)#get_info_cpu(multi._outf('log'),mulslice=multi.is_mulslice)
            df.loc[name][info_cols] = info
            for file in files:multi.ssh_get(ssh_host,file)
    df.to_pickle(df_path);
    print(green+'DataFrame updated and saved : \n' +yellow+df_path+black)
    return df

#########################################################################
### def : display
#########################################################################
def plot_beam_thickness(beams,rip='I',linespec='-',marker='x',cm='Greens',**kwargs):
    ''' plot the beams as function of thickness
    - beams : [hk,t,re,im,Ib] beams info from import_beams or file containing them
    - Ib : nbeams x nthicknesss
    - rip flags : I(Intens),r(real),i(imag)
    '''
    hk,t,re,im,Ib = beams
    nbs = len(hk)
    csp,csr,csi,plts = dsp.getCs(cm,nbs),dsp.getCs('Blues',nbs),dsp.getCs('Reds',nbs),[]
    for i in range(nbs):
        if 'I' in rip : plts += [[t,Ib[i],[csp[i],'%s%s' %(linespec,marker)],'$I_{%s}$' %(hk[i])]]
        if 'r' in rip : plts += [[t,re[i],csr[i],'$re$']]
        if 'i' in rip : plts += [[t,im[i],csi[i],'$im$']]
    return dsp.stddisp(plts,lw=2,labs=['$thickness(\AA)$','$I_{hk}$'],**kwargs)#,xylims=[0,t.max(),0,5])

def plot_fe(qmax=1,Nx=20,Vcell=1,**kwargs):
    q0,fq_e = scat.get_elec_atomic_factors(['Si'],q=np.linspace(0,qmax,1000))
    vg = fq_e[0]/Vcell
    vgmax = vg.max()
    Xq =np.arange(Nx)[:,None].dot(np.array([1,1])[None,:]).T/ax
    Yq = np.tile([0,vgmax],[Nx,1]).T
    plts=[[Xq,Yq,[(0.7,1.0,0.7),'--'],''],
          [q0,vg,'g--','$v_g(Si)$']]
    dsp.stddisp(plts,labs=['$q(A^{-1})$','$v_g(q)(A^{-2})$'], xylims=[0,qmax,0,vgmax],
        legLoc='upper right',**kwargs)

def plot_pattern_section(file='si_autoslic100.tif'):
    im=plt.imread(datpath+file)
    N=im.shape[0]
    std.stddisp([range(N),im[:,0],'b','$I_{h0}$'])




class Multi_Pattern_viewer():
    def __init__(self,multi,patterns,figpath,i=0,**args):
        ''' View cbf files
        - exp_path : path to images
        - figpath : place to save the figures
        - i : starting image
        '''
        self.multi = multi
        self.figpath = figpath
        self.patterns  = np.sort(patterns);#print(patterns)
        self.args = args
        self.nfigs = self.patterns.size
        self.fig,self.ax = dsp.stddisp() #caxis=[0,1],pOpt='im')#,caxis=[0,0.001])
        cid = self.fig.canvas.mpl_connect('key_press_event', self)
        self.i=i     #starting image
        self.inc=1   #increment(use 'p' or 'm' to change)
        self.mode=1
        # rc('text', usetex=False)
        # self.multi.pattern(fig=self.fig,ax=self.ax,file=self.patterns[self.i],
        #     title='pattern %d' %self.i,**self.args)
        self.update()

    def update(self):
        self.ax.cla()
        print("%d/%d" %(self.i,self.nfigs))
        zi = self.multi.i_slice*self.multi.slice_thick*(self.i+1)
        tle = 'z=%d A' %(zi)
        # tle = r'%s' %dsp.basename(self.patterns[self.i]).replace('_',' ')
        self.multi.pattern(fig=self.fig,ax=self.ax,file=self.patterns[self.i],
            title=tle,opt='',pOpt='tX',imOpt='',**self.args)
        self.fig.canvas.draw()

    def __call__(self, event):
        # print(event.key)
        if event.key in ['up','right']:
            self.i=min(self.i+self.inc,self.nfigs-1)
            self.mode=1
            self.update()
        elif event.key in ['left','down']:
            self.i=max(0,self.i-self.inc)
            self.mode=-1
            self.update()

        if event.key=='s':
            dsp.saveFig(self.figpath+'pattern%s.png' %str(self.i).zfill(3),ax=self.ax)

        if event.key=='p':
            self.inc=min(self.inc+1,100);print(self.inc)
        if event.key=='m':
            self.inc=max(1,self.inc-1);print(self.inc)



###########################################################################
#### def : test
###########################################################################
if __name__ == "__main__":
    plt.close('all')



import numpy as np
import os

f = np.load(os.path.dirname(__file__)+'/data/abcd.npy')

def Va(r,Z,ip=np.arange(3)) :
    al,ag=150.4121417,266.5985798;
    V=np.zeros(r.shape)
    for i in ip:
        ai,bi = f[Z][2*i:2*i+2]
        ci,di = f[Z][6+2*i:6+2*i+2]
        # print(ai,bi)
        # print(ci,di)
        fi = ai/r*np.exp(-2*np.pi*r*np.sqrt(bi))
        gi = ci*di**(-3/2)*np.exp(-np.pi**2*r**2/di)
        V+=al*fi + ag*gi
    return V



"""Bloch contiuous rotation experiment"""
import importlib as imp
import os,glob,numpy as np
from subprocess import Popen,check_output
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from utils import displayStandards as dsp#;imp.reload(dsp)
from utils import glob_colors as colors#;imp.reload(dsp)
from EDutils import display as EDdisp #;imp.reload(EDdisp)
from EDutils import rotate_exp        ;imp.reload(rotate_exp)
from EDutils import utilities as ut   #;imp.reload(ut)
from EDutils import viewers as vw     #;imp.reload(vw)
from . import bloch                   ;imp.reload(bloch)
from . import util as bu              ;imp.reload(bu)



class Bloch_cont(rotate_exp.Rocking):
    def __init__(self,**kwargs):
        """ Bloch continous rocking curve

        Parameters
        ----------
        kwargs
            to pass to :class:~EDutils.rotate_exp.Rocking
        """
        super().__init__(bloch.Bloch,**kwargs)
        self.figpath = os.path.join(self.path,'tiff')
        # self.thick=
        if not os.path.exists(self.figpath):
            Popen('mkdir %s' %self.figpath,shell=True)
        self.save()

    def set_beams_vs_thickness(self,thicks,v=1):
        self.do('_set_beams_vs_thickness',thicks=thicks,v=v)
        self.z = self.load(0).z

    def sum_images(self,n,fmt='',figpath=None,frames=()):
        ''' Sums images found in figpath by chuncks of n images
        and puts them into directory "figpath/sum"

        Parameters
        ----------
        n
            number of images to sum (each side of the central image will be n/2 images)
        figpath
            Where the images are located
        frames
            The range of frames consider in (f_init,f_end)
        '''
        #handle output style
        if not fmt:fmt = glob.glob(os.path.join(figpath,'*'))[0].split('.')[-1]
        sum_path=os.path.join(figpath,'sum')
        pad_n = int(np.ceil(np.log10(np.ceil(self.df.shape[0]/n))))
        if not os.path.exists(sum_path):
            Popen('mkdir %s' %sum_path,shell=True)

        # handles min and max frames
        nmax = self.df.shape[0]
        n_max = int(np.ceil(nmax/n))
        if not len(frames)==2:frames=(0,n_max+1)
        n_init,n_end = frames
        n_init,n_end=max(0,n_init),min(n_end,n_max)
        n2 = n//2 #int(np.ceil(n/2))

        ## check images exists
        nbounds=lambda n0:max(0,min(nmax-1,n0))
        ni,nf = nbounds(n*n_init-n2),nbounds(n*n_end+n2+1)
        filenames = np.array([figpath+'/%s.%s' %(self.load(i).name,fmt)
            for i in [ni,nf]])
        miss = [not os.path.exists(f) for f in filenames]
        if any(miss):
            print(colors.red+'Missing images : \n'+colors.black)
            print('\n'.join(filenames[miss]))
            return
        #get the size of the images
        nxy = bu.imread(filenames[0]).shape
        for i in np.arange(n_init,n_end+1):
            im = np.zeros(nxy)
            subframes = np.arange(max(0,i*n-n2),min(nmax,i*n+n2+1))
            print(colors.red,i,subframes,colors.black)
            for j in subframes:
                b = self.load(j)
                img_file=os.path.join(figpath,'%s.%s' %(b.name,fmt))
                im+=bu.imread(img_file)/n
            frame_str = str(i).zfill(pad_n)
            new_file = os.path.join(sum_path,'%s.%s' %(frame_str,fmt))
            out = check_output("cp %s %s" %(filenames[0],new_file),shell=True).decode()
            if out:print(out)
            bu.imwrite(new_file,im.T)#np.fliplr(np.flipud(im)))
            # print(colors.yellow+new_file+colors.green+' saved'+colors.black)

    def make_img(self,template=None,figpath=None,fmt='',frames=None,**kwargs):
        if not figpath:figpath=self.figpath
        nmax=self.df.shape[0]
        if template and not fmt:fmt=template.split('.')[-1]
        if type(frames)==type(None):frames=np.arange(nmax)
        if not os.path.exists(figpath):
            out=check_output('mkdir -p %s' %figpath,shell=True).decode()
            if out:print(out)
        for i in frames:
            b0 = self.load(i)
            filename=figpath+'/%s.%s' %(b0.name,fmt)
            b0.convert2img(filename,template,**kwargs)


    def convert2tiff(self,figpath=None,n=0,nmax=0,**kwargs):
        ''' Generate tiff files
        Parameters
        ------------
            figpath
                final name will be figpath+'/%s.tiff' %b.name
            kwargs
                passed to Bloch.convert2tiff
            n
                if n> 1 sum n images at a time
            nmax
                number of frames to consider
        '''
        import tifffile
        if not figpath:figpath=self.figpath
        if not nmax:nmax=self.df.shape[0]
        if n>1:
            b0=self.load(0)
            tiff_file=figpath+'/%s.tiff' %b0.name
            im=tifffile.imread(tiff_file)
            sum_path=os.path.join(figpath,'sum')
            # self.sum_path=sum_path
            pad_n = int(np.ceil(np.log10(np.ceil(self.df.shape[0]/n))))
            if not os.path.exists(sum_path):
                Popen('mkdir %s' %sum_path,shell=True)

            for i in range(0,nmax,n):
                im = np.zeros(im.shape)
                new_tiff_file = os.path.join(sum_path,'%s.tiff' %(str(i//n).zfill(pad_n)))
                # for j in range(n):
                for j in range(max(0,i+n-5),min(nmax-1,i+n+5)):
                    b = self.load(j)
                    tiff_file=figpath+'/%s.tiff' %b.name
                    im+=tifffile.imread(tiff_file)/n
                tifffile.imwrite(new_tiff_file,im)
                print(colors.yellow+new_tiff_file+colors.green+' saved'+colors.black)

        else:
            for i in range(nmax):
                b = self.load(i)
                b.convert2tiff(tiff_file=figpath+'/%s.tiff' %b.name,**kwargs)
                b.save()

    def convert2png(self,cutoff=20,n=None,**kwargs):
        import tifffile
        for i in range(self.df[:n].shape[0]):
            b = self.load(i)
            tiff_file=self.figpath+'/%s.tiff' %b.name
            im=tifffile.imread(tiff_file)
            dsp.stddisp(im=[im],
                cmap='gray',caxis=[0,cutoff],axPos='F',pOpt='p',
                opt='sc',name=tiff_file.replace('.tiff','.png'),figsize='im',**kwargs)

    def show_tiff(self,sum_opt=False,**kwargs):
        # figpath=self.figpath
        sum_path=os.path.join(self.figpath,'sum')
        if sum_opt:figpath=sum_path
        return vw.Base_Viewer(figpath,**kwargs)

    def plot_rocking(self,cond='',refl=[],**kwargs):
        if not any(refl) and not cond:cond=strong_beams
        return super().plot_rocking(cond=cond,refl=refl,**kwargs)

def strong_beams(dfG,
        tol:float=1e-2,n:int=10,
        opt:str='F'):
    """keep the n strongest beams, i.e. those with intensity I>tol not including central beam

    Parameters
    ----------
    dfG
        the bloch dataframe
    tol
        strong beam criteria
    n
        number of strong beams
    opt
        include Origin(O) Friedel pairs(F)
    Returns
    ----------
    x
        list of beams
    """
    dfM=dfG.copy()
    if not 'O' in opt:dfM = dfM.drop(str((0,0,0)))
    dfM = dfM.sort_values('I',ascending=False)[:2*n]
    if not 'F' in opt:remove_friedel_pairs(dfM)

    df = dfM.loc[dfM.I>tol]
    if df.shape[0]>n:df = df[:n]
    if df.shape[0]==0:df= dfM[:n]
    # print(df.shape,n)

    # print(dfM)
    return list(df.index.values)




# def bloch_rock(tag,path='',opts='',
#     uvw=None,u=None,u1=None,omega=np.linspace(-1,1,3),
#     thicks=(0,1000,1000),bloch_args={},hkls=None,
#     S_args={},W_args={},R_args={},Q_args={},I_args={},Z_args={},
#     refl=[],cond='',ts0=None,i=0,iZs=None,zs=[],cm='hsv',
#     opt='p'):
#     ''' complete a rocking curve simulation sweep with Blochwave approach
#     - tag : Sweep tag
#     - uvw,u,u1,omega : beam orientation vectors : see ut.get_uvw_rock
#     - bloch_args : thicknesses
#     - opts : options to perform
#         - generic : s(solve all), t(set_thickness),
#         - ts0 dependent       : S(setup), I(Iz)
#         - zs dependent        : R(rocking curve),Q(Idyn-Ikin plot)
#         - refl dependent only : W(Excitation error),Z(integrated)
#     - refl,cond : selected beams
#     - iZs,zs : thicknesses to select (RQ)
#     - i,ts0  : single simulation to select for opts(SWI)
#     '''
#     b_args = {}#'keV':200,'solve':1,'opts':'s','thick':thicks[1]}
#     b_args.update(bloch_args)
#
#     pkl = os.path.join(path,'rock_%s.pkl' %tag)
#     figpath = os.path.join(path,'figures')
#     if 's' in opts:
#         if not isinstance(uvw,np.ndarray):
#             uvw  = ut.get_uvw_rock(u0=u,u1=u1,omega=omega)
#         rock = exp.Rocking(Simu=bloch.Bloch,param='u',vals=uvw,ts=omega,
#             tag=tag,path=path,thicks=thicks,**b_args)
#     else:
#         rock = ut.load_pkl(pkl)
#
#     if 't' in opts:
#         rock.do('set_beams_vs_thickness',thicks=thicks)
#
#     if any([c in opts for c in 'XSI']):
#         i,ts0 = rock._get_ts(i,ts0)
#         b = rock.load(i=i)#;print(b.u)
#
#     for ch in opts:
#         if   ch=='X' :print(b.get_Xig())
#         elif ch=='S' :
#             figname = '%s_Sw.svg' %tag
#             idx = b.get_beam(cond=cond,refl=refl)#;print(idx)
#             frame_args = {'opts':'SVkr','mag':500,'xylims':1.5,'rot':0}
#             frame_args.update(S_args)
#             EDdisp.show_frame(df_bloch=b.df_G,single_mode=False,
#                 hkl_idx=idx,name=os.path.join(figpath,figname),opt=opt,
#                 **frame_args)
#         elif ch=='I':
#             figname = '%s_Iz.svg' %tag
#             refl += [[0,0,0]]
#             b.show_beams_vs_thickness(thicks=thicks,refl=refl,cond=cond,
#                 title=r'$\theta=%.2f$' %ts0,cm='Spectral',
#                 name=os.path.join(figpath,figname),opt=opt,
#                 **I_args)
#         elif ch=='W':
#             figname = '%s_theta' %tag
#             fz = lambda x:-np.log10(np.maximum(x,1e-10))
#             sw_args = {'cm':cm,'fz':fz,'opts':'t'}#,'thick':thick}
#             sw_args.update(W_args)
#             rock.Sw_vs_theta(refl=refl,cond=cond,
#                 figname=os.path.join(figpath,figname),opt=opt,lw=2,
#                 **sw_args)
#         elif ch=='R':
#             rocking_args={'cmap':'Spectral'}
#             if not isinstance(hkls,list):hkls = [refl]
#             for iB,hkl in enumerate(hkls):
#                 figname = '%s_beams%d.svg' %(tag,iB)
#                 rocking_args.update(R_args)
#                 rock.plot_rocking(zs=zs,cond='',refl=hkl,
#                     lw=2,opt=opt,name=os.path.join(figpath,figname),
#                     **rocking_args)
#         elif ch=='Q':
#             figname = '%s_QQ.svg' %tag
#             rock.QQplot(zs=zs,iZs=iZs,refl=refl,cond=cond,
#                 lw=2,opt=opt,name=os.path.join(figpath,figname),
#                 **Q_args)
#         elif ch=='Z':
#             figname = '%s_Iint.svg' %tag
#             int_args = {'cm':cm}
#             int_args.update(Z_args)
#             rock.plot_integrated(cond=cond,refl=refl,
#                 lw=2,opt=opt,name=os.path.join(figpath,figname),
#                 **int_args)
#
#     return rock
#
#
#
#
# def load_pkl(file):
#     with open(file,'rb') as f : obj = pickle5.load(f)
#     return obj



from utils import*
uni = colors.unicolor
Zs = [1,6,7,8,14,16]
cs   = [uni(0.9),uni(0.5),'b','r','m','y']
legs = ['H','C','N','O','Si','S']

nRmax = 100
fspline ='data/splines.npy'

def convert():
    nZs = len(Zs)
    ss = (nZs,nRmax)
    y,b,c,d = np.zeros(ss),np.zeros(ss),np.zeros(ss),np.zeros(ss)
    for iz,iZ in enumerate(Zs):
        x,y[iz],b[iz],c[iz],d[iz] = np.loadtxt('data/spline_%d.txt' %iZ).T
    np.save(fspline,[x,y,b,c,d])
    print(colors.green +'saved : '+colors.yellow+fspline+colors.black)

def fv_spline(x0,iz):
    i = np.argmin(abs(x0-x))
    z = x0 - x[i];
    vz = y[iz][i] + ( b[iz][i] + ( c[iz][i] + d[iz][i] *z ) *z) *z;
    return vz


def show_Vz():
    r = np.linspace(0,5,1000)

    plts = [ [r,np.array([fv_spline(r0**2,iz) for r0 in r]), cs[iz],legs[iz] ] for iz,iZ in enumerate(Zs)]
    plts = plts[4]
    dsp.stddisp(plts,labs=['r','Vz'],lw=2,xylims=[0,3,0,500])


# convert()
x,y,b,c,d = np.load(fspline,allow_pickle=True)
show_Vz()



'''
A Multislice interface to build decks, launch jobs and postprocess multislice simulations\n
[TOC]\n
##Example
Create a periodic simulation from data in 'Silicon/Si110',
submit it and monitor its state :
```python
multi=Multislice('Silicon/Si110',mulslice=False,NxNy=512,repeat=[6,6,40],Nhk=2)
p = multi.run()
multi.check_simu_state()
```
Once terminated, get the beams intensities as function of thickness,
forcing to repostprocess in case it was already done in an older simulation :
```python
multi.beam_vs_thickness(bOpt='f')
```
'''
import importlib as imp
import pickle,socket,time,tifffile
import pandas as pd
import numpy as np
from math import ceil,nan
from subprocess import Popen,check_output,PIPE
from glob import glob as lsfiles
from os.path import dirname,realpath,exists,basename
from string import ascii_uppercase as SLICES #ascii_letters
from string import ascii_letters
# from matplotlib.animation import FuncAnimation
# from utils.glob_colors import*
from crystals import Crystal
import utils.glob_colors as colors
import utils.displayStandards as dsp                        ;imp.reload(dsp)
from . import postprocess as pp                             ;imp.reload(pp)
from . import mupy_utils as mut                             ;imp.reload(mut)
from .config import ssh_hosts,temsim_hosts,path_hosts  

class Multislice:
    ''' **DATA PARAMETERS :**\n
    - `name` : path to the simulation folder
    - `tail` : trailing string in the naming convention of outputs (tag is an alias as well)
        - name pattern = *name*\_*tail*\_*program_name*
    - `data` : data file names
        - if None and mulslice simu : all *[a-zA-Z].dat and stacked alphabetically
        - if None and autoslic simu : first *.xyz found
        - Otherwise paths to filename(s) of the data files to use (must be in name)
    \n**MULTISLICE simulation parameters :**\n
    - `mulslice` : False(autoslic), True(mulslice)=>periodic
    - `keV` : float - wavelength in keV
    - `tilt` : 2-list - crystal tilt tx,ty (mrad)
    - `repeat` : 3-int-list - super cell size in the x,y,z directions
    - `NxNy` : 2-int-list or int : sampling in x and y (same sampling in x and y if only int provided)
    - `slice_thick` : float - slice thickness (A) (`dz` can also be used as an alias)
    - `i_slice` : record intensity every i_slice
    - `hk` : list of tuple - beams to record as function of depth
    - `Nhk`: int - set hk on a grid of multiples of the supercell size. Prevails over hk argument if >0. default(0)
    - `hk_sym` : True (take the Friedel pairs for the beams in hk)
    \n**Thermal parameters**:\n
    - `TDS` : bool - include thermal vibrations (as dictated by wobble parameter in .dat )
    - `T`   : Temperature in K if TDS is True
    - `nTDS` : Number of runs to average over
    \n**RUN/OUTPUT options:**\n
    - `v`  : verbose - str or bool(all if True) or int(verbose level)
        - n(naming pattern),c(cell params),t(thickness),r(run cmd)
        - d(data),D(Decks),R(full run cmd)
    - `opt` : d(save_deck), s(save_obj), r(do_run),  f(force rerun),w(ask before running again), p(do_pp)
    - `fopt` : If the simulation was previously completed :\n
        - '' : The simulation will not be run again
        - 'w' is on  (default case) : the user will be asked to confirm whether he wants to run it again
        - 'f' in fopt : The simulation is rerun
    - `ppopt` : u(update),w(wait), I(image) B(beam) P(pattern) A(azim_avg) S(reduced pattern) s(save_patterns in job)
    - `ssh` : ip address or alias of the machine on which to run the job
    - `cif_file` : name of .cif file corresponding to structure in path
    ### OBSOLETE
    - `hostpath` : path to data on host
    - `cluster` : indicate if running on a cluster using qsub
    '''
    def __init__(self,name,mulslice=False,tail='',tag=None,data=None,u=None,
                tilt=[0,0],TDS=False,T=300,n_TDS=16,
                keV=200,repeat=[1,1,1],NxNy=[512,512],slice_thick=1.0,dz=None,i_slice=1000,
                hk=[(0,0)],Nhk=0,hk_pad=None,hk_sym=0,prev=None,
                opt='sr',fopt='',ppopt='uwPs',v=1,
                ssh=None,hostpath='',cluster=False,cif_file=None,xyz_params=None):

        v = self._get_verbose_options(v)
        #attributes
        self.version     = '1.4.3b'
        self.name        = dsp.basename(name)                       #;print('basename:',self.name)
        self.datpath     = realpath(name)+'/'                       #;print('path:',self.datpath)
        self.cif_file    = self.get_cif_file(cif_file)              #;print(self.cif_file)
        self.is_mulslice = mulslice
        self.tilt        = tuple(tilt)
        self.TDS         = TDS
        self.T           = T
        self.n_TDS       = n_TDS
        self.tail        = self._set_tail(tail,tag)
        self.name        = self._set_name('n' in v)
        self.data        = self._get_datafiles(data)
        self.u           = u
        self.xyz_params  = xyz_params
        self.slices      = SLICES[:len(self.data)]
        self.outf        = self._set_filenames()
        self.keV         = keV
        self.repeat      = tuple(repeat)
        self.NxNy        = self._set_NxNy(NxNy)
        self.hk          = self._set_hk(hk,Nhk,hk_sym,hk_pad)
        self.cell_params = self._get_cell_params('c' in v)
        self.slice_thick = self._set_slice_thick(slice_thick,dz)
        self.i_slice     = i_slice
        self.thickness   = self._get_thickness('t' in v)
        self.decks       = {}
        self.p           = None

        ## make decks and run if required
        self.execute(opt,fopt,ppopt,v, ssh,cluster,hostpath,prev)

    ########################################################################
    ##### Public functions
    ########################################################################
    def make_decks(self,save=True,datpath=None,prev=None,v=False):
        '''create the decks from the information provided'''
        if self.is_mulslice:
            decks={self.outf['simu_deck']:self._mulslice_deck(prev)}
            for i in self.slices:
                decks[self.outf['slice_deck%s' %i]]=self._atompot_deck(i)
        else:
            decks={self.outf['simu_deck']:self._autoslic_deck(prev)}
        if save:
            #datapath is manually provided for remote generation
            if not datpath : datpath = self.datpath
            if v:print(colors.green+'Decks saved :'+colors.black)
            for filename,deck in decks.items():
                with open(datpath+filename,'w') as f : f.write(deck)
                if v:print(colors.yellow+datpath+filename+colors.black)
        self.decks = decks

    def save(self,v=False):
        '''save this multislice object'''
        file=self._outf('obj')
        with open(file,'wb') as out :
            pickle.dump(self, out, pickle.HIGHEST_PROTOCOL)
        if v:print(colors.green+"object saved\n"+colors.yellow+file+colors.black)

    def execute(self,opt='sr',fopt='',ppopt='w',v=1, ssh='',cluster=False,hostpath='',prev=None):
        # save_deck,save_obj,do_run,do_pp,fopt,vopt = self._get_run_options(opt,fopt,v)
        save_deck,save_obj,do_run,do_pp = [s in opt for s in 'dsrp']
        save_deck |= do_run
        if 'f' in opt: fopt += 'f'
        if 'w' in opt: fopt += 'w'
        vopt=('r' in v and 'R' not in v) + 2*('R' in v)
        self.patterns_saved,patterns_opt=0,'s' in ppopt         #patterns_saved set to 1 during job postprocess
        if save_deck : self.make_decks(save_deck,prev=prev,v=v)
        if save_obj : self.save(v=v)
        if 'd' in v : self.print_datafiles()
        if 'D' in v : self.print_decks()
        if do_run : self.p = self.run(v=vopt, fopt=fopt,ssh_alias=ssh,hostpath=hostpath,cluster=cluster,patterns_opt=patterns_opt)
        if do_pp : self.postprocess(ppopt,ssh,hostpath=hostpath)
        self._set_figpath()

    def resume(self,Nz,v=1,opt='sr',
        hk=None,Nhk=None,hk_sym=0,i_slice=None,prev=1,
        **kwargs):
        ''' resume a simulation from its last point
        - Nz : number of unit cells to run
        - hk,Nhk,hk_sym,i_slice : values which can be set
        - **kwargs : opt,fopt,ppopt,v, ssh,cluster,hostpath
        '''
        v = self._get_verbose_options(v=v)
        prev = self._set_prev(prev)
        self.repeat = self.repeat[:2]+(Nz,)
        if hk or Nhk : self.hk = self._set_hk(hk,Nhk,hk_sym)
        if i_slice : self.i_slice = i_slice
        self.thickness += self._get_thickness('t' in v)
        self.execute(v=v,prev=prev,opt='sr'+opt,**kwargs)
        self.merged=0
        self.merge_beams()

    def run(self,v=1,fopt='w',ssh_alias=None,hostpath=None,cluster=False,patterns_opt=False,):
        '''run the simulation with temsim
        - fopt : f(force rerun), w(warn ask rerun already done)
        - ssh : name of the host to run the job
        '''
        if isinstance(fopt,int):fopt='f'
        run = True
        if self.check_simu_state(v=0,ssh_alias=''):
            if v>0 : print(colors.red+"Simulation already performed in the past."+colors.black)
            if 'f' in fopt :
                msg='Force re-running'
                cmd = '''echo Deleting logfile and patterns
                if [ -f %s ];then rm %s;fi
                rm %s
                '''%(self._outf('log'),self._outf('log'),self._outf('pattern')+'*')
                p = Popen(cmd, shell=True,stderr=PIPE,stdout=PIPE)
                p.wait();p.communicate()
            else :
                if 'w' in fopt :
                    run = input(colors.red+"Re-run?(y/n) : "+colors.black)=='y'
                    msg = 're-running'
                else :
                    run,msg = False, 'not running'
            if v : print(colors.red+msg+colors.black)
        p = None
        if run:
            if ssh_alias :
                if ssh_alias=='badb':cluster=1
                cmd = self._get_job_ssh(ssh_alias,hostpath=hostpath,v=v>2,cluster=cluster,patterns_opt=patterns_opt)
                #time.sleep(1)
                #self.check_simu_state(v=0,ssh_alias=ssh_alias,hostpath=hostpath)
            else :
                self._get_job(cluster=cluster, patterns_opt=patterns_opt)
                cmd = 'bash %s' %self._outf('job')
            p = Popen(cmd,shell=True) #; print(cmd)
            if v>0 : print(colors.green+self.name+" job submitted at %s" %time.ctime()+colors.black)
            if v>1 : print(colors.magenta+cmd+colors.black)
        return p

    def postprocess(self,ppopt='wuP',ssh_alias='',tol=1e-4,figpath=None,opt='p',hostpath=''):
        '''Performs postprocessing with predefined options : \n
        - ppopt:w(wait until done) u(update) d(display) for I(image), B(beam) P(pattern)     #f(force recalculate)
        - figpath : Directory to place the figures if saving with automatic naming (default datpath)
        '''
        print(colors.blue+'...postprocessing...'+colors.black)
        if not figpath : figpath = self.datpath
        time.sleep(0.1)
        state = self.check_simu_state(ssh_alias=ssh_alias,hostpath=hostpath);print(state)
        if not state == 'done' and 'w' in ppopt:
            self.wait_simu(ssh_alias,10,hostpath)
            self.p.wait()
        # self.get_beams(iBs='a',bOpt='f')
        if ssh_alias and 'u' in ppopt:
            if 'I' in ppopt : self.ssh_get(ssh_alias,'image'     ,hostpath)
            if 'B' in ppopt : self.ssh_get(ssh_alias,'beams'     ,hostpath)
            if 'P' in ppopt : self.ssh_get(ssh_alias,'patternnpy',hostpath)
            if 'S' in ppopt : self.ssh_get(ssh_alias,'patternS'  ,hostpath)
        #convert to np.array
        if 'd' in ppopt:
            if 'I' in ppopt and opt : self.image(opt=opt,name=figpath+self.outf['imagesvg'])
            if 'B' in ppopt and opt :
                # if not exists(self._outf('beams')) or 'f' in ppopt:self.get_beams(bOpt='fa')
                self.beam_vs_thickness(bOpt='f',tol=tol,opt=opt,name=figpath+self.outf['beamssvg'])
            if 'P' in ppopt and opt :
                # if not exists(self._outf('patternnpy')) or 'f' in ppopt:self.save_pattern()
                self.pattern(Iopt='Incsl',tol=tol,imOpt='ch',cmap='gray',opt=opt,name=figpath+self.outf['patternsvg'])

    ###################################################################
    ################ OUTPUT FUNCTIONS
    ###################################################################
    def get_cif_file(self,cif_file):
        if not cif_file:cif_file=self.name+'.cif'
        return self.datpath+cif_file
    def get_structure_factor(self,**sf_args):
        '''computes structure factor 3D
        - sf_args : see (mut.get_structure_factor3D)
        returns :
        - (qx,qy,qz),Fhkl
        '''
        return mut.get_structure_factor(self.cif_file,**sf_args)
        # crys = Crystal.from_cif(self.cif_file)
        # pattern = np.array([np.hstack([a.coords_fractional,a.atomic_number]) for a in crys.atoms] )
        # lat_vec = np.array(crys.reciprocal_vectors)
        # (h,k,l),Fhkl = structure_factor3D(pattern, lat_vec, **sf_args)
        # qx = h/crys.lattice_parameters[0]
        # qy = k/crys.lattice_parameters[1]
        # qz = l/crys.lattice_parameters[2]
        # return (qx,qy,qz),Fhkl

    def image(self,opt='I',cmap='jet',**kwargs):
        '''Displays the 2D image out of simulation
        - opt : I(intensity)
        '''
        im=plt.imread(self._outf('image'))
        if 'I' in opt :
            N = int(im.shape[1]/2)
            real,imag = im[:,:N],im[:,N:];#print(real.max())
            im = real**2+imag**2
            im =im/im.max()
        dsp.stddisp(labs=['$x$','$y$'],im=im,legOpt=0,imOpt='c',**kwargs)

    def gif_beams(self,step=5,cmap='gray'):
        '''make a gif of the saved beams pattern'''
        hk,ts,re,im,Ib = self.get_beams(iBs=[],tol=1e-5,bOpt='fa')
        ax,by = self.cell_params[:2]
        Nx,Ny,Nz = self.repeat
        h,k = np.array(self.hk).T
        h,k = h/ax/Nx,k/by/Ny
        # nzf=int(np.log10(Nz/step))+1
        Ib = np.log10(Ib)
        Imax=Ib.max()
        gif_path=self.datpath+'gif/'#+self.name
        p=Popen("if [ ! -d %s ];then mkdir %s;fi" %(gif_path,gif_path),shell=True);p.wait()#.decode()
        ##### display
        # fig,ax = dsp.create_fig(figsize='f')
        # ims = []
        for i,it in enumerate(ts[::step]):
            tle = '$z=%.0f \AA$' %it
            # im = ax.scatter(h,k,Ib[:,i],cmap=cmap,vmin=-1,vmax=,animated=True)
            # ims = [im]
            dsp.stddisp(scat=[h,k,Ib[:,i]],labs=['$q_x(\AA^{-1})$','$q_y(\AA^{-1})$'],
                caxis=[0,Imax],imOpt='ch',#xylims=[-5,5,-5,5],
                texts=[0,4,tle,'g'],
                name=gif_path+'%s.png' %str(i).zfill(nzf),opt='sc')

        # dsp.standardDisplay(labs=['$q_x(\AA^{-1})$','$q_y(\AA^{-1})$'],
        #     ,caxis=[0,Imax],imOpt='ch',#xylims=[-5,5,-5,5],
        #     texts=[0,4,tle,'g'],
        #     name=gif_path+'%s.png' %str(i).zfill(nzf),opt='sc')

        # ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True,
        #     repeat_delay=1000)
        # ani.save('%s/beams.mp4' %self.datpath)
        # fig.show()

        # cmd = "im2gif %s && eog %s.gif" %(gif_path,gif_path)
        # cmd="convert -delay 20 -loop 0 %s*.png %s" %(gif_path,self.datpath+self.name+'.gif')
        # print("creating the gif")
        # print(check_output(cmd,shell=True).decode())

    def beam_vs_thickness(self,bOpt='',orig=0,iBs=[],tol=1e-2,**kwargs):
        ''' plot beam vs thickness
        - bOpt : O(include Origin),p(print Imax),n or f(force new)
        - iBs : beams (recorded beams are found in self.hk)
            - list of str ['(h0,k0)','(h1,k1)'..]
            - indices
        - kwargs : see help(plot_beam_thickness)
        '''
        hk,t,re,im,Ib = np.load(self._outf('beams'),allow_pickle=True)
        # hk,t,re,im,Ib = self.get_beams(iBs=[],tol=1e-5,bOpt='fa')
        # print(hk)
        if orig:bOpt+='O'
        if any(iBs) :
            if isinstance(iBs[0],str) :
                iBs = [ (iB==hk).argmax() for iB in iBs if (iB==hk).sum()]
                print(iBs)
            if isinstance(iBs[0],tuple):
                iBs = [i for i,hk0 in enumerate(self.hk) if hk0 in iBs]
        else :
            Imax = np.array([I.max() for I in Ib]) #;print(Imax)
            iBs = [i for i in range('O' not in bOpt,len(hk)) if Imax[i]>= tol*Imax.max()]
        beams = np.array(hk)[iBs],t, np.array(re)[iBs],np.array(im)[iBs],np.array(Ib)[iBs]

        if 'o' in bOpt:
            return beams
        else:
            return pp.plot_beam_thickness(beams,**kwargs)

    def merge_beams(self):
        if not self.merged:
            beamnpy = self._outf('beams')
            p = Popen('cp %s %s' %(beamnpy,beamnpy+'%s' %str(int(self.thickness)).zfill(4)),stderr=PIPE,stdout=PIPE,shell=True)
            p.wait();p.communicate()
            beams_files = np.sort(lsfiles(self._outf('beams')+'*'))[1:]     #; print(beams_files)
            hk,t,re,im,I  = np.load(beams_files[0],allow_pickle=True)
            z=t.max()
            for file in beams_files[1:]:
                hk_i,t_i,re_i,im_i,I_i = np.load(file,allow_pickle=True)
                t_i +=z                                                     #;print(file,z)
                t,I = np.hstack([t,t_i]),np.hstack([I,I_i])
                z   += t_i.max()
            # self.merged=1
            np.save(beamnpy,[hk,t,re,im,I])
            print(colors.green+'beams file merged : \n'+colors.yellow+beamnpy+colors.black)
        else:
            print(colors.pruple + 'beams already merged' + colors.black)

    def get_beams(self,iBs=[],tol=1e-2,bOpt=''):
        ''' get the beams as recorded during:\n
        - iBs : selected beam indices : default=Ibeams.max()>tol
        - bOpt : O(include Origin),p(print Imax),n or f(force new)
        hk,t,re,im,Ib = beams
        '''
        if 'a' in bOpt : iBs = range(len(self.hk))
        new = 'n' in bOpt or 'f' in bOpt #;print(bOpt)
        if not exists(self._outf('beams')) or new:
            beams = pp.import_beams(self._outf('beamstxt'),self.slice_thick,iBs,tol,'O' in bOpt,'p' in bOpt)
            np.save(self._outf('beams'),beams)
            print(colors.green+'beams file saved : \n'+colors.yellow+self._outf('beams')+colors.black)
        beams = np.load(self._outf('beams'),allow_pickle=True)
        return beams

    def save_patterns(self,force=0,save_opt=1,v=0):
        if force : self.patterns_saved=0
        if not self.patterns_saved:
            print(colors.blue+'...saving patterns to npy...'+colors.black)
            i_files = np.sort([f.replace(self._outf('pattern'),'') for f in lsfiles(self._outf('pattern')+'*')])
            for i in i_files :
                self.save_pattern(i=i,v=v)
            self.patterns_saved=1
            if save_opt:self.save(v=1)
    def save_pattern(self,iz=None,i='',v=1):
        if iz:
            patterns = np.sort(lsfiles(self._outf('pattern')+'*'))
            txt_file = patterns[min(patterns.size-1,iz+1)]
            i=str(iz).zfill(3)
        else:
            txt_file=self._outf('pattern')+i
        if v:print('loading pattern %s' %txt_file)
        im = np.loadtxt(txt_file)
        real,imag = im[:,0:-1:2],im[:,1::2];#print(real.max())
        im = real**2+imag**2
        npy_file=self._outf('pattern').replace('.txt','')+i.replace('.','')+'.npy'
        np.save(npy_file,im,allow_pickle=True)
        if v:print(colors.green+'file saved : ' +colors.yellow+npy_file+colors.black)

    def show_patterns(self,**kwargs):
        patterns = lsfiles(self._outf('pattern').replace('.txt','')+'0*.npy')
        return pp.Multi_Pattern_viewer(self,patterns,figpath=self.datpath,**kwargs)

    # def _get_patterns(self):return

    def patterns2gif(self,name=None,v=0,**kwargs):
        if not name:
            self._set_figpath()
            name=self.figpath+self.name+'_pattern.gif'
        self.save_patterns(v=0)

        name = name.replace('.gif','')
        patterns = np.sort(lsfiles(self._outf('pattern').replace('.txt','')+'*.npy'))[1:]
        print(colors.blue+'...saving patterns to png...'+colors.black)
        for iz in range(len(patterns)):
            figname='%s%s.png' %(name,str(iz).zfill(4))
            self.pattern(iz=iz,name=figname,opt='sc',v=v,**kwargs)
        print(colors.blue+'...saving patterns to gif...'+colors.black)
        cmd='im2gif '+name +' png'
        out=check_output(['/bin/bash','-i','-c',cmd]).decode()
        print(colors.green+out+colors.black)

    def set_thicks(self):
        self.dzs = self.i_slice*self.slice_thick
        self.zs = np.arange(0,self.thickness,self.dzs)

    def get_iz(self,thick,v=0):
        self.dzs = self.i_slice*self.slice_thick
        self.zs = np.arange(self.dzs,self.thickness+self.slice_thick,self.dzs)

        iz = np.argmin(abs(self.zs-thick))
        if v==1: print('thick=%.2f, actual thick=%.2f' %(thick,self.zs[iz]))
        if v==2:return iz,self.zs[iz]
        return iz

    def integrate_reflections(self,idxs,dq=None,N=10,**kwargs):
        self.set_thicks()
        if isinstance(N,int):N=[N]*2
        if dq :
            dqxy = self.repeat[:2]/self.cell_params[:2]
            dqx,dqy = Nh/ax,Nk/by
            N = np.round(dq/dqxy)
        Nx,Ny = np.array(N,dtype=int)
        nbs = idxs.shape[0]
        zs = self.zs.copy()#[:20]
        self.bs = np.zeros((nbs,zs.size))
        for iz,z in enumerate(zs):
            qx,qy,im = self.pattern(iz=iz,out=1,**kwargs)
            for i,idx in enumerate(idxs):
                s = np.ix_(idx[0]+np.arange(-Nx,Nx+1),idx[1]+np.arange(-Ny,Ny+1))
                # print(im[s].sum())
                self.bs[i,iz] = im[s].sum()

        cs = dsp.getCs('jet',nbs)
        plts = [[zs,b,cs[iB],'%d' %iB] for iB,b in enumerate(self.bs)]
        dsp.stddisp(plts,labs=['z','Iint'])

    def pattern(self,iz=None,file=None,rmax=10,Iopt='Ncs',out=0,tol=1e-6,Nmax=None,gs=3,Imax=3e4,
        rot=0,rings=[],v=1,cmap='binary',pOpt='im',title='',name=None,**kwargs):
        '''Displays the 2D diffraction pattern out of simulation
        - Iopt : t(tiff), c(crop I[r]<tol), n(normalize), s(fftshift), l(logscale), q(quarter only) r(rand) g(good)
        - Nmax : crop display beyond Nmax
        - rings : list or array - of resolutions for rings to display
        - gs    : broadening range
        - rmax  : indicative of noise level fall off
        - rot   : rotate pattern anticlock by rot
        - kwargs : see stddisp
        returns : [qx,qy,I]
        '''
        if isinstance(iz,int):
            npy_files = self._outf('patternnpy').replace('.npy','[0-9]*.npy')
            patterns = np.sort(lsfiles(npy_files))#;print(patterns)
            izs = np.array([p.split('pattern')[-1].replace('.npy','') for p in patterns],dtype=int)
            # print(patterns)
            idx = np.where(izs-iz==0)[0]#;print(idx)
            if idx.size:
                file = patterns[idx[0]]#;print(file)
                zi = self.i_slice*self.slice_thick*(iz+1)#;print(zi)
        if not file:
            file = self._outf('patternnpy')
            if not exists(file):file=self._outf('patternS')#;print('ok')
            zi = self.thickness
        if not title:title = 'z=%d A' %(zi)

        if v:print('loading %s at z=%.1fA' %(file,zi))
        im = np.load(file)
        if v>1:print('original image shape',im.shape)
        ax,by = self.cell_params[:2]
        Nh,Nk = self.repeat[:2];
        Nx,Ny = np.array(np.array(self.NxNy)/2,dtype=int)
        NMAX = 1024
        if Nmax : Nmax = min(Nmax,Nx,Ny)
        # if 'I' in Iopt :
            # real,imag = im[:,0:-1:2],im[:,1::2];#print(real.max())
            # im = real**2+imag**2

            #print('normalizing')
        if 'N' in Iopt:
            im/=(4*Nx*Ny)**2
            if v>1:print(im.max(),im.sum())
        elif 'n' in Iopt : #normalize:
            im00 = im[0,0];im[0,0] = 0
            mMax = im.max();im /= mMax
            if 'l' in Iopt :
                im[0,0] = im00/mMax
            else:
                im[0,0] = 1

        if 'c' in Iopt and not Nmax:
            #print('croping')
            idx = im[:Nx,:Ny]>tol#*im.max()
            h,k = np.meshgrid(np.arange(Nx),np.arange(Ny))
            r = np.sqrt(h**2+k**2)
            Nmax = ceil(r[idx].max()); #print('Pattern Nmax=%d > %E ' %(Nmax,tol))
            Nmax = min(Nmax,NMAX)#;print(Nx,Ny,Nmax)
        if not Nmax : Nmax = min(NMAX,Nx,Ny)

        if 's' in Iopt : im = np.fft.fftshift(im)   #;print('fftshift')

        #print('preparing')
        if 'q' in Iopt:
            im0 = im[Nx:Nx+Nmax,Ny:Ny+Nmax];del(im)
            h,k = np.meshgrid(np.arange(Nmax),np.arange(Nmax))
        else:
            im0 = im[Nx-Nmax:Nx+Nmax,Ny-Nmax:Ny+Nmax];del(im)#;print(im0.shape)
            h,k = np.meshgrid(np.arange(-Nmax,Nmax),np.arange(-Nmax,Nmax))

        if 'g' in Iopt:
            # print(np.where(im0>10*tol))
            i,j = np.where(im0>10*tol) #index of spots
            # print(i,j)
            im00 = im0.copy()
        if 'g' in Iopt:
            gs3 = gs
            dqx,dqy = Nh/ax,Nk/by
            nx,ny = int(np.floor(gs3/dqx)),int(np.floor(gs3/dqy)) ;
            if v>1: print('Gaussian window function size : ', nx,ny)
            ix,iy = np.meshgrid(range(-nx,nx+1),range(-ny,ny+1))
            x,y = ix*dqx,iy*dqy
            Pb = np.exp(-(x**2+y**2)/(gs3/3)**2)#;dsp.stddisp(im=[x,y,Pb],pOpt='im')
            # Pb = 1/(x**2+y**2+0.01)**10
            # print((Pb/Pb.sum()).max())
            for i0,j0 in zip(i,j):
                i0x,j0y = i0+ix,j0+iy #;print(i0x,j0y)
                idx         =  (i0x>=0) & (j0y>=0)  & (i0x<2*Nmax) & (j0y<2*Nmax)
                i0x,j0y     = i0+ix[idx],j0+iy[idx]
                # print(idx)#i0x,j0y)
                im0[i0,j0] -= im00[i0,j0]
                im0[i0x,j0y] += Pb[idx]/Pb[idx].sum()*im00[i0,j0]
                # im0[i0,j0] = 100#im00
                # im0[i0x,j0y] = np.maximum(im0[i0x,j0y],tol*1e-2)

        if 'r' in Iopt :
            r = np.sqrt(h**2+k**2);r[r==0]=1
            im0 += np.random.rand(im0.shape[0],im0.shape[1])/(rmax*r)

        if 'l' in Iopt : #logscale the data
            im0[im0<tol] = tol*1e-2
            im0 = np.log10(im0)

        qx,qy = h/Nh/ax,k/Nk/by
        if rot:
            alpha = np.deg2rad(rot)
            ct,st = np.cos(alpha),np.sin(alpha)
            qx,qy = ct*qx-st*qy,st*qx+ct*qy
        N = [1,4]['q' in Iopt]
        if out : return qx,qy,np.array(im0*Imax,dtype='uint16')

        self._set_figpath()
        if not name:name=self.figpath+basename(file).replace('.npy','.png')

        t = np.linspace(0,2*np.pi/N,100)
        ct,st = np.cos(t),np.sin(t)
        plts = [[r*ct,r*st,'g--',''] for r in rings]
        if v:print('displaying pattern:',im0.shape)

        fig,ax = dsp.stddisp(plts,labs=[r'$q_x(\AA^{-1})$','$q_y(\AA^{-1})$'],im=[qx,qy,im0],
            cmap=cmap,pOpt=pOpt,title=title,name=name,**kwargs)
        if 't' in Iopt:
            tiff_file = name.replace('.png','.tiff')
            self.convert2tiff(tiff_file,im0,Nmax,rot,n=256,Imax=Imax)
        return fig,ax

    def convert2tiff(self,tiff_file,im0,n0,rot,n=256,Imax=5e4):
        alpha = np.deg2rad(rot)
        ct,st = np.cos(alpha),np.sin(alpha)
        u_n   = np.arange(-n,n+1)
        hi,ki = np.meshgrid(u_n,u_n)
        Hi,Ki = ct*hi+st*ki,st*hi-ct*ki
        H,K = np.array(np.round(Hi)+n0,dtype=int),np.array(np.round(Ki)+n0,dtype=int)
        # dsp.stddisp(im=[I[H,K]],caxis=[0,50],pOpt='im',cmap='gray')

        I = np.array(im0*Imax,dtype='uint16')
        tifffile.imwrite(tiff_file,I[H,K]) #,np.flipud(I))
        print(colors.yellow+tiff_file+colors.green+' saved'+colors.black)

    def azim_avg(self,tol=1e-6,Iopt='Incsl',out=0,**kwargs):
        ''' Display the average azimuthal diffraction pattern intensities
        - out : get data only
        '''
        qx,qy,I = self.pattern(out=1,opt='',tol=tol,Iopt=Iopt)
        #brute force
        print('brute force averaging')
        qr = np.round(np.sqrt(qx**2+qy**2)*1e10)/1e10
        qr0 = np.unique(qr)#,return_index=1,return_inverse=1)
        I0 = np.array([ I[np.where(qr==q)].mean() for q in qr0])
        if out:
            return qr0,I0
        else:
            plts = [[qr0,I0,'b-','']]
            dsp.stddisp(plts,labs=[r'$q(\AA^{-1})$','$I_q$'],**kwargs)

    ########################################################################
    ###### print functions
    ########################################################################
    def print_datafiles(self,data=None):
        '''show data files *.dat or .xyz depending on mulslice'''
        if not data : data = self.data
        #if not isinstance(data,list) : data = [data]
        self._print_header('%s FILES' %(['.xyz','*.dat'][self.is_mulslice]))
        for dat in data :
            print(colors.green+dat+colors.black)
            with open(self.datpath+dat,'r') as f: print(''.join(f.readlines()))
            print("\n")

    def print_decks(self,decks=None):
        '''show decks *.in'''
        self._print_header('*.in FILES')
        if not decks:decks=list(self.decks.keys())
        if not isinstance(decks,list) : decks = [decks]
        for deck in decks:
            print(colors.green+deck+colors.black)
            try:
                with open(self.datpath+deck,'r') as f: print(''.join(f.readlines()))
            except FileNotFoundError:
                print(self.decks[deck])
                print(colors.red+"WARNING:deck file %s not on disk" %(self.datpath+deck)+colors.black)
    def print_job(self):
        '''print bash script job file '''
        with open(self._outf('job'),'r') as f:
            self._print_header('.sh FILE')
            print(''.join(f.readlines()))
    def print_log(self,head_opt=0):
        '''print the logfile of running multislice .log'''
        try:
            with open(self._outf('log'),'r') as f:
                self._print_header('.log FILE')
                log = f.readlines()
            if head_opt:
                for i,l in enumerate(log):
                    if  'Sorting atoms' in l :break
                print(''.join(log[:i+1]))
            else:
                print(''.join(log))
        except FileNotFoundError:
            print(colors.red+'logfile not created yet, run a simu first'+colors.black)
            return 0
    def log_info(self,v=1):
        with open(self._outf('log'),'r') as f : lines=f.readlines()
        if self.is_mulslice:
            log_lines = [log_lines[-6]] + log_lines[-2:]
            info = [l.strip().split('=')[-1].split(' ')[1] for l in log_lines]
            info = [log_lines[0].strip().split(',')[0].replace('slice','')]+info #zmax
        else:
            state = self.check_simu_state()
            zmax,Imax,cpuT,wallT='0','1','0','0'
            if not state=='init':
                zl = [i for i,l in enumerate(lines) if 'z=' in l ][-1]
                zmax,Imax =[l.strip().split('=')[-1].split(' ')[1] for l in lines[zl:zl+2]]
            if state=='done':
                wl = [i for i,l in enumerate(lines) if 'wall time' in l ][0]
                cpuT,wallT = [l.strip().split('=')[-1].split(' ')[1] for l in lines[wl-1:wl+1]]
        info = np.array([zmax,Imax,cpuT,wallT],dtype=float)
        if v:
            # print(info,log_lines)
            print('zmax=%.1f,Imax=%.4f,cpuT=%.1f,wallT=%.1f' %tuple(info))
        return info

    def wait_simu(self,ssh_alias='',t=1,hostpath=''):
        state = 0
        while not state=='done':
            state=self.check_simu_state(ssh_alias,v=0,hostpath=hostpath)
            print(state)
            time.sleep(t)

    def check_simu_state(self,ssh_alias=None,v=False,hostpath=''):
        '''see completion state of a simulation '''
        if ssh_alias :
            e = self.ssh_get(ssh_alias,'log',hostpath=hostpath)
            if e:
                print(colors.red+e+colors.black)
                if 'No such file' in e:return 'not started'
        try:
            with open(self._outf('log'),'r') as f :
                lines = f.readlines()  #print(self._outf('log')) #f.readlines())
                if len(lines)>=2 :
                    l1,l2=lines[-2:];#print(l1,l2)
                else :
                    return 'empty'
            # get state
            if self.is_mulslice:
                if 'slice ' in l2 :
                    slice       = int(l2.split(',')[0].split(' ')[-1]);
                    tot_slice   = len(self.data)*self.repeat[2]
                    state="%d%%" %(100*slice/tot_slice)
                elif 'elapsed time' in l2 : state='done'
                else : state='init'
            else:
                zl = [i for i,l in enumerate(lines) if 'z=' in l ]
                if not sum(['Sorting atoms' in l for l in lines]) : state='init'
                elif 'END' in l2 : state='done'
                elif sum(['wall time' in l for l in lines]) : state='processing'
                elif len(zl) :
                    l1=lines[zl[-1]]
                    state="%d%%" %int(100*float(l1[3:8])/self.thickness)
                else:
                    state='undefined'
            #print state info
            if v :
                if state=='init': print(colors.green+"Initializing..."+colors.black)
                elif state=='done' : print(colors.green+"Done : \n"+colors.black, l1,l2)
                else : print(colors.green+state.ljust(5)+black )
            return state
        except FileNotFoundError:
            print(colors.red+'logfile not created yet, run a simu first'+colors.black)
            return ''

    ########################################################################
    ######## Private functions
    ########################################################################
    def _get_verbose_options(self,v):
        if isinstance(v,bool) : v=['','nctrdD'][v]
        if isinstance(v,int) : v=''.join(['nctr','DR','d'][:min(v,3)])
        return v
    # def _get_run_options(self,opt,fopt):
    #     save_deck,save_obj,do_run,do_pp = [s in opt for s in 'dsrp']
    #     save_deck |= do_run
    #     if 'f' in opt: fopt += 'f'
    #     if 'w' in opt: fopt += 'w'
    #     vopt=('r' in v and 'R' not in v) + 2*('R' in v)
    #     return save_deck,save_obj,do_run,do_pp,fopt,vopt

    def _get_datafiles(self,data):
        dat_type = ['xyz','dat'][self.is_mulslice]
        err_msg=colors.red+'no *.' +dat_type+' files found in :\n'+colors.yellow+self.datpath+colors.black
        if not data : data = lsfiles(self.datpath+'*.'+dat_type)
        if not data : raise Exception(err_msg)
        if not isinstance(data,list) : data=[data]; #print(data[0].split('.')[-1])
        if not data[0].split('.')[-1]==dat_type : print(err_msg)# raise Exception(err_msg)
        data = [dsp.basename(f) for f in data ]
        return data

    def _set_filenames(self):
        outf = {'obj'       : self.name+'.pkl',
                'simu_deck' : self.name+'.in',
                'log'       : self.name+'.log',
                'image'     : self.name+'.tif',
                'imagesvg'  : self.name+'.svg',
                'pattern'   : self.name+'_pattern.txt',
                'patternnpy': self.name+'_pattern.npy',
                'patternS'  : self.name+'_patternS.npy',
                'patternsvg': self.name+'_pattern.svg',
                'beamstxt'  : self.name+'_beams.txt',
                'beams'     : self.name+'_beams.npy',
                'beamssvg'  : self.name+'_beams.svg',
                'job'       : self.name+'.sh',
                }
        if self.is_mulslice:
            for dat,i in zip(self.data,self.slices):
                outf['slice_data%s' %i] = dat
                outf['slice_imag%s' %i] = dat.replace('.dat','.tif')
                outf['slice_deck%s' %i] = dat.replace('.dat',self.tail+'.in')
        else :
            outf['data'] = self.data[0]
        return outf

    def _set_figpath(self):
        self.figpath = self.datpath+'figures/'
        if not exists(self.figpath):
            Popen('mkdir %s ' %(self.figpath),shell=True)

    def _get_cell_params(self,v=False):
        if self.is_mulslice :
            cz=[]
            for i in list(self.slices):
                with open(self._outf('slice_data%s' %i)) as f:
                    ax,by,cz_i = np.array(f.readline().rstrip().split(' '),dtype=float)
                    cz.append(cz_i)
            cell_params = [ax,by,cz]
        else :
            with open(self._outf('data')) as f:
                f.readline()
                ax,by,cz = np.array(f.readline().rstrip().split(' '),dtype=float)
                cell_params = [ax,by,[cz]]
        if v : print('ax=%.3fA, by=%.3f, cz=' %(ax,by), cz)
        return cell_params

    def _set_tail(self,tail,tag=None):
        if isinstance(tag,str):tail=tag
        tail = '_'+tail if tail else ''
        return tail
    def _set_name(self,v=False):
        name = self.name+self.tail+['_autoslic','_mulslice'][self.is_mulslice]
        if v : print('Simu name pattern = '+name)
        return name
    def _set_NxNy(self,NxNy):
        if isinstance(NxNy,int) or  isinstance(NxNy,np.int64) :
            NxNy = [NxNy,NxNy]
        return tuple(NxNy)
    def _set_hk(self,hk,Nhk,sym=0,hk_pad=None):
        Nh,Nk,Nl=self.repeat
        if hk_pad:
            if isinstance(hk_pad,int):hk_pad=[hk_pad]*2
            Nh,Nk=hk_pad
        if Nhk:
            h,k = np.meshgrid(range(-sym*Nhk,Nhk),range(-sym*Nhk,Nhk))
            hk=[(Nh*h,Nk*k) for h,k in zip(h.flatten(),k.flatten())];#print(hk)
        self.Nhk=Nhk
        self.hk_pad=hk_pad
        return hk
    def _set_slice_thick(self,slice_thick,dz=None):
        if dz:slice_thick=dz
        if self.is_mulslice : slice_thick = self.cell_params[2]
        return slice_thick
    def _get_thickness(self,v=True):
        cz = np.array(self.cell_params[2]).sum()
        thickness = self.repeat[2]*cz
        if v : print('simulated thickness = %.3f A, nslices=%d' %(thickness,thickness/self.slice_thick))
        return thickness
    def _set_prev(self,prev):
        zstr = str(int(self.thickness)).zfill(4)
        if isinstance(prev,bool) or isinstance(prev,int):
            if prev:
                prev = self.outf['image']+zstr
        if prev:
            files = ['log', 'patternnpy','patternS','image','beams']
            cmd = ''
            for ext in files:
                file = self._outf(ext)
                old_file = file + zstr
                cmd += "if [ -f %s ]; then cp %s %s;fi\n" %(file,file,old_file)
            # files = list(self.outf.keys());files.remove('obj');files.remove('data');files.remove('beamstxt')
            # for ext in files:
            #     file = self._outf(ext)
            #     cmd += "if [ -f %s ]; then rm %s;fi\n" %(file,file)
            p = Popen(cmd,shell=True,stderr=PIPE,stdout=PIPE)
            p.wait();p.communicate()
        return prev

    ########################################################################
    #### Job
    ########################################################################
    def _job_pp(self,patterns_opt):
        datpath = self.datpath;
        self.datpath='./';
        try:
            self.get_beams(bOpt='fa');
        except:
            print('get_beams failed');
        try:
            self.save_pattern();
            qx,qy,It1 = self.pattern(Iopt='Ncs',out=True,Nmax=260);
            np.save(self.outf['patternS'],[qx,qy,It1]);
        except:
            print('pattern failed')
        if patterns_opt:
            self.save_patterns(save_opt=0)
            self.datpath = datpath
            self.save()

    def _get_job(self,temsim=None,cluster=False,datpath=None,patterns_opt=False):
        logfile     = self.outf['log'] #self._outf('log')
        simu_deck   = self.outf['simu_deck'] #self._outf('simu_deck')
        prog        = ['autoslic','mulslice'][self.is_mulslice]
        if not temsim : temsim = temsim_hosts[socket.gethostname()]
        header = self._version_header()
        #write the cmd
        job = "#!/bin/bash\n"
        if cluster:job += "#$ -j y\n#$ -cwd\n#$ -V -w e\n"
        job += '\ncd %s \n' %self.datpath
        job += 'printf "%s" > %s \n\n' %(header,logfile) #overwrite logfile
        if self.is_mulslice:
            for i in self.slices :
                deck = self.outf['slice_deck%s' %i]
                job += 'cat %s | %s >> %s \n' %(deck,temsim+'atompot',logfile)
        job += 'cat %s | %s >> %s\n' %(simu_deck,temsim+prog,logfile)

        #### postprocess
        pyexe='python3'
        if cluster:pyexe='/home/lii26466/anaconda3/bin/python3'
        # import numpy as np;
        # beams = pp.import_beams('%s',%s);
        # np.save('%s',beams)
        # ''' %(self._outf('beamstxt'),self.slice_thick,self._outf('beams'))
        job+='printf "\n\tPOSTPROCESSING\n"  >> %s \n\n' %(logfile)
        pycode='''import multislice.postprocess as pp;import numpy as np;
        multi = pp.load_multi_obj('%s');
        multi._job_pp(%d)
        ''' %(self.outf['obj'],patterns_opt)

        job+='%s -c "%s" >>%s 2>&1 \n' %(pyexe, pycode.replace('\n',''),logfile)
        job+='printf "END" >>%s\n' %(logfile)

        #save job
        if not datpath:datpath=self.datpath
        with open(datpath+self.outf['job'],'w') as f : f.write(job)
        print(colors.red+colors.yellow+datpath+self.outf['job']+colors.black)

    def _get_job_ssh(self,ssh_alias,hostpath=None,v=False,cluster=False,patterns_opt=False):
        #save local datpath
        datpath  = self.datpath
        hostpath = self._get_hostpath(ssh_alias,hostpath)
        self.datpath = hostpath

        #save updated deck and job
        host    = ssh_hosts[ssh_alias]
        temsim  = temsim_hosts[host]
        self._get_job(temsim,cluster,datpath,patterns_opt=patterns_opt)
        self.make_decks(save=True,datpath=datpath)

        #create directory on remote if does not exist
        cmd = 'ssh %s "if [ ! -d %s ]; then mkdir -p %s;echo %s:%s created;fi"' %(ssh_alias,hostpath,hostpath,ssh_alias,hostpath)
        print(colors.blue+check_output(cmd,shell=True).decode()+colors.black)
        cmd = 'ssh %s "if [ -f %s ]; then  echo 1;fi"' %(ssh_alias,hostpath+self.data[0])
        # dat_exist = 0
        dat_exist = check_output(cmd,shell=True).decode()#;print(hostpath+self.data[0],dat_exist)
        #copy files over to remote
        cmd  = 'cd %s && scp ' %(datpath)
        if not dat_exist:cmd += '%s ' %(' '.join(self.data))
        cmd += '%s ' %(' '.join(list(self.decks.keys())))
        cmd += '%s ' %(self.outf['job'])
        cmd += '%s ' %(self.outf['obj'])
        cmd += '%s:%s' %(ssh_alias,hostpath)
        # check_output(cmd,shell=True).decode()
        p = Popen(cmd,shell=True);p.wait()

        # submit job command (qsub for clusters)
        if cluster:
            cmd = 'source /etc/bashrc && qsub -S /bin/bash '
        else :
            cmd = 'bash '
        cmd += '%s' %(hostpath+self.outf['job'])
        cmd = 'ssh %s "%s"' %(ssh_alias,cmd)
        #restore
        self.datpath = datpath
        self.make_decks(save=True)
        return cmd

    def ssh_get(self,ssh_alias,file,hostpath=None,dest_path=None):
        if not dest_path : dest_path = self.datpath
        hostpath = self._get_hostpath(ssh_alias,hostpath)
        cmd = 'scp %s:%s %s' %(ssh_alias,hostpath+self.outf[file],dest_path)
        p = Popen(cmd,shell=True,stderr=PIPE)#stdout=PIPE

        p.wait()
        o,e = p.communicate()
        # print(o.decode())
        return e.decode()

    ########################################################################
    #### decks
    ########################################################################
    def _atompot_deck(self,i):
        # TODO (prevent mulslice from adding the extension automatically for this case )
        dat_file = self._outf('slice_data%s' %i)
        img_file = self._outf('slice_imag%s' %i).replace('.tif','')
        deck  = "%s\n" %dat_file                #.dat file
        deck += "%s\n" %img_file                #.tif file
        deck += "n\n"                           #record fft proj
        deck += "%d %d\n" %(self.NxNy)          #sampling
        deck += "%d %d 1\n" %(self.repeat[:2])  #super cell
        deck += "n\n"                           #thermal displacement
        return deck

    def _mulslice_deck(self,prev=None):
        prev_run = "%s\n" %['y','n'][prev==None]
        if prev : prev_run += "%s\n" %(self.datpath+prev)
        Nz,sl = self.repeat[2],ascii_letters[:len(self.data)]
        deck = "%d(%s)\n" %(Nz,sl)              #Stack sequence
        for i in self.slices :                  #atom pots
            deck+="%s\n" %self._outf('slice_imag%s' %i)#
        deck += "%s\n" %(self._outf('image'))   #image file
        deck += "n\n"                           #partial coherence
        deck += prev_run                        #start previous run
        deck += "%.2f\n" %self.keV              #wavelength
        deck += "0 0\n"                         #crystal tilt
        deck += "%.4f %.4f \n" %self.tilt       #beam tilt
        deck += "y\n"                           #record beams
        deck += "%s\n" %(self._outf('beamstxt'))#     filename
        deck += "%d\n" %len(self.hk)            #     nbeams
        for hk in self.hk :                     #     beam indices
            deck += "%d %d\n" %(hk)
        #deck += "n\n"                           #record cross sections
        #deck +=
        return deck

    def _autoslic_deck(self,prev=None):
        prev_run = "%s\n" %['y','n'][prev==None]
        if prev : prev_run += "%s\n" %(self.datpath+prev)
        #Deck
        deck  = "%s\n" %self._outf('data')      #.xyz file
        deck += "%d %d %d\n" %self.repeat       #super cell
        deck += "%s\n" %self._outf('image')     #image file
        deck += "n\n"                           #partial coherence
        deck += prev_run                        #start previous run
        if not prev:
            deck += "%.4f\n" %self.keV              #wavelength
            deck += "%d %d\n" %self.NxNy            #sampling
        deck += "%.4f %.4f \n" %self.tilt       #beam tilt
        deck += "%f\n" %(self.slice_thick)      #slice thickness
        deck += "y\n"                           #record beams
        deck += "%s\n" %self._outf('beamstxt')  #filename
        deck += "%d\n" %len(self.hk)            #   nbeams
        for hk in self.hk :                     #   - beams -
            deck += "%d %d\n" %(hk)             #   beam indices
        deck += "%s\n" %(['n','y'][self.TDS])   #thermal vibration?
        if self.TDS:                            #   - Params -
            deck+='%.3f\n' %self.T              #   temperature
            deck+="%d\n" %self.n_TDS            #   nb configurations
        deck += "n\n"                           #intensity cross section
        deck += "y\n"                           #diffraction pattern
        deck += "%s\n" %self._outf('pattern')   #   filename
        deck += "%d" %self.i_slice;
        return deck

    ########################################################################
    #### misc private
    def _version_header(self):
        header = '''
------------------------------------------------------------------------------
------------------------------------------------------------------------------
This header has been produced by multislice library
version : %s
date : %s
author : Tarik Ronan Drevon
e-mail : tarik.drevon@stfc.ac.uk
------------------------------------------------------------------------------
------------------------------------------------------------------------------
''' %(self.version,time.ctime())
        return header
    def _print_header(self,msg,w=70):
        head='#'*w
        print(colors.blue+head)
        print('\t\t\t' +msg+ ' :')
        print(head+colors.black)

    def _get_hostpath(self,ssh_alias,hostpath):
        if not hostpath :
            local_hostpath  = path_hosts[socket.gethostname()]
            remote_hostpath = path_hosts[ssh_hosts[ssh_alias]]
            hostpath  = self.datpath.replace(local_hostpath,remote_hostpath)
            # simu_folder = self.datpath.split('/')[-2]
            # hostpath += simu_folder+'/'
            # hostpath = self.datpath.replace('ronan','tarik').replace('CCP4','git/ccp4')
        return hostpath

    def _outf(self,file):
        return self.datpath+self.outf[file]

#########################################################################
##### utilities
#########################################################################
def get_tilts(tx=np.arange(0,10,0.05),ty=0):
    def convert_tilts(tilts,ntilts):
        if isinstance(tilts,float) or isinstance(tilts,int):
            tilts = [tilts]*ntilts
        return np.deg2rad(tilts)*1000

    if isinstance(tx,list) : tx=np.array(tx)
    if isinstance(ty,list) : ty=np.array(ty)
    nts = 0
    if isinstance(tx,np.ndarray):
        nts=tx.size
    if isinstance(ty,np.ndarray):
        if nts:
            if not nts==ty.size:
                raise Exception('tx and ty have different sizes')
        nts = ty.size
    if not nts:
        raise Exception('at list one array of tilts must be provided ')
    txs = convert_tilts(tx,nts)
    tys = convert_tilts(ty,nts)
    tilts = [[tx,ty] for tx,ty in zip(txs,tys)]
    return tilts

class Rocking:
    def __init__(self,name,tx=np.arange(0,10,0.05),ty=0,tag='',**kwargs):
        ''' simulate rocking curve
        - tx : tilt parameters around x(degrees)
            - float - constant value
            - list or np.ndarray : actual list of tilts
        - ty : tilt parameters around y(same behaviour as tx)
        - tag : tag will then be '<tag>_tilt<nb>'
        '''
        if tag:tag+='_'
        self.path = name
        self.tag  = tag
        self.df_path = self.path+self.tag+'tilts.pkl'
        self.tx,self.ty = tx,ty
        self.tilts = get_tilts(tx,ty)
        self.df = sweep_var(name,param='tilt',vals=self.tilts,tail=self.tag,df=self.tag+'tilts.pkl',
            **kwargs)
        self.save(v=1)

    def save(self,v=0):
        '''save this object'''
        file = self.path+self.tag+'rock.pkl'
        with open(file,'wb') as out :
            pickle.dump(self, out, pickle.HIGHEST_PROTOCOL)
        if v:print(colors.green+"object saved\n"+colors.yellow+file+colors.black)


    #############################################################################
    # utilities
    #############################################################################
    def load(self,i):
        return pp.load_multi_obj(self.path+self.df.index[i])

    def update(self,files=[],v=1):
        df = pp.update_df_info(self.df_path,files)
        if v:
            np.set_printoptions(precision=3)
            pd.set_option("precision", 2)
            print(df[['tilt','state','zmax(A)','Inorm']])

    def plot_rocking(self,iBs,iZs=-1,zs=None,**kwargs):
        '''plot rocking curve for beams iBs at thickness zs
        - iBs : list - beam indices [1,2,5] or ids [(1,0),(1,1)]
        - iZs : int or list - slice indices (last slice by default )
        - zs : float or list - selected thickness (in A) to show(takes preference over iZs if set)
        '''
        nbs,nzs,iZs = self._init_rocking(iBs,iZs,zs)

        I = np.zeros((len(self.tilts),nbs,nzs))
        for i,pkl in enumerate(self.df.index):
            multi = pp.load_multi_obj(self.path+pkl)
            hk,z,re,im,Ib = multi.beam_vs_thickness(bOpt='o',iBs=iBs)           #;print(Ib.shape)
            I[i] = np.array(Ib[:,iZs])

        ts = self.tx
        cs,ms,plts = dsp.getCs('jet',nbs), dsp.markers,[]
        legElt = { '%s' %hk[i]:[cs[i],'-'] for i,iB in enumerate(iBs)}
        for iz,iZ in enumerate(z[iZs]):
            legElt.update({'$z=%d A$' %(iZ):['k',ms[iz]+'-']})
            plts += [[ts,I[:,i,iz],[cs[i],ms[iz]+'-'],''] for i,iB in enumerate(iBs)]
        dsp.stddisp(plts,labs=[r'$\theta$(deg)','$I$'],legElt=legElt,**kwargs)

    #############################################################################
    #### misc
    #############################################################################
    def _init_rocking(self,iBs,iZs,zs):
        multi = pp.load_multi_obj(self.path+self.df.index[0])
        hk,z,re,im,Ib = multi.beam_vs_thickness(bOpt='o',iBs=iBs)
        if isinstance(zs,float) or isinstance(zs,float):zs = [zs]
        if isinstance(zs,list) or isinstance(zs,np.ndarray):
            iZs = [np.argmin(abs(z-z0)) for z0 in zs]
        if isinstance(iZs,int):iZs=[iZs]
        nbs = np.array(hk).size
        nzs = z[iZs].size
        return nbs,nzs,iZs

def sweep_var(name,param,vals,df=1,ssh='',tail='',do_prev=0,**kwargs):
    '''runs a set of similar simulations with one varying parameter
    - name          : path to the simulation folder
    - param,vals    : the parameters and values to sweep
    - df            :
        - pd.Dataframe to update(since parsed as a reference)
        - int create and save the new dataframe if 1
    - do_prev       : Used for iterative fourier transform
    - kwargs : see help(Multislice)
    '''
    do_df,save = isinstance(df,pd.core.frame.DataFrame),0
    dfname = 'df.pkl'
    if isinstance(df,str):dfname,df=df,1
    if isinstance(df,int):
        if df : df,do_df,save = pd.DataFrame(columns=[param,'host','state']+pp.info_cols),1,1
    nvals,prev = len(vals),None
    for i,val in zip(range(nvals),vals):
        print(colors.red+param+':',val,colors.black)
        kwargs[param]=val
        if do_prev and i: prev = multi.outf['image']
        multi=Multislice(name,prev=prev,
            ssh=ssh,tail=tail+param+str(i).zfill(ceil(nvals/10)),
            **kwargs)
        if do_df:
            df.loc[multi.outf['obj']] = [nan]*len(df.columns)
            df.loc[multi.outf['obj']][[param,'host','state']] = [val,ssh,'start']
    if save :
        df.to_pickle(name+dfname)
        print(colors.green+'Dataframe saved : '+colors.yellow+name+dfname+colors.black)
    return df



import importlib as imp
import pickle,matplotlib
import numpy as np, pandas as pd
import scipy.fftpack as fft
from scipy.integrate import nquad,trapz,quad
import utils.displayStandards as dsp
import utils.physicsConstants as cst
import utils.glob_colors as colors
import wallpp.lattice as lat
import wallpp.plane_group as pg     #;imp.reload(pg)
from . import pymultislice;imp.reload(pymultislice)
# from scipy.signal import fftconvolve
# from scipy.interpolate import interp1d
# import utils.FourierUtils as fu


class Multi2D(pymultislice.Multislice):
    '''multislice 2D for quick testing
    - pattern : list of 2d-arrays - [x0,z0,f] where x,z=np.meshgrid(x0,z0)
    - ax,bz : lattice constants
    - **kwargs : see pymultislice.Multislice.__init__
    '''
    def __init__(self,pattern,ax,bz,**kwargs):
        self.version = 2.0
        self.ax = ax
        self.bz = bz
        self.ez = self.bz
        print(colors.red+'\t\t 2D multislice simulation '+colors.black)
        super().__init__(pattern,**kwargs)

    def run(self,nz,iZs,iZv,opts,v):
        if self.TDS:
            self._TDS(nz,iZs,iZv,opts,v)
        else:
            self._std_mutlislice(nz,iZs,iZv,opts,v)

    def _TDS(self,nz,iZs,iZv,opts,v):
        pattern  = np.vstack([self.pattern + np.array([self.ax*i,0,0]) for i in range(self.Nx)])
        wobbles  = self.wobble[np.array(pattern[:,-1],dtype=int)][:,None]
        pattern0 = pattern
        Na = pattern0.shape[0]
        self.patterns=[]
        for iTDS in range(self.nTDS):
            print(colors.yellow+'configuration %d' %iTDS+ colors.black)
            pattern[:,:2] = pattern0[:,:2]+(2*np.random.rand(Na,2)-1)*wobbles#*[self.ax,self.bz]
            self.patterns+=[pattern.copy()]
            p1 = pg.Wallpaper('p1',self.Nx*self.ax,self.bz,90,pattern,ndeg=self.nx,gen=True)
            self.pattern = p1.get_potential_grid_p1()

            if v:print(colors.blue+'...integrating projected potential...'+colors.black)
            self._set_transmission_function()
            self._set_propagator()
            self.set_Psi0(iTDS)
            self.propagate(nz,iZs,iZv,opts,iTDS,v)
        self.psi_qz=np.abs(self.psi_qz)**2
        self.pattern=pattern0

    def _std_mutlislice(self,nz,iZs,iZv,opts,v):
        self._set_transmission_function()
        self._set_propagator()
        self.set_Psi0()
        if nz:self.propagate(nz,iZs,iZv,opts,iTDS=0,v=v)

    ###################################################################
    #### display
    ###################################################################
    def V_show(self,**kwargs):
        x0,z0,fv = self.pattern
        x,z = np.meshgrid(x0,z0)

        args = {'imOpt':'cv','axPos':'V','xylims':[z0.min(),z0.max(),x0.min(),x0.max()]}
        args.update(kwargs)
        dsp.stddisp(im=[z,x,fv],labs=['$z$','$x$'],**args)

    def Za_show(self):
        atoms = pd.DataFrame.from_dict(
            { 0:[(0,0,0),15], 1:[(0.5,0.5,0.5),40],
              2:[(0,0,1),45], 3:[(1,0,0),50],
              4:[(0,1,1),80]},
            orient='index',columns=['c','s'])

        markers = list(matplotlib.markers.MarkerStyle.markers.keys())[2:-4]
        nms=len(markers)
        if self.TDS:
            scat = ()
            for i,p in enumerate(self.patterns):
                Za = atoms.iloc[p[:,2]]
                # print(Za.size)
                scat += ([p[:,0],p[:,1],Za.s,Za.c,markers[i%nms]] ,)
            dsp.stddisp(scat=scat,labs=['$x$','$z$'])
        else:
            print('pattern with atomic positions need be provided')

    def Tz_show(self,iSz=slice(0,None,1),opts='',Vopt='VT',cmaps=['Greens','Blues','Reds'],**kwargs):
        '''Show Transmission function
        - opts or Vopt : 'V(potential) 'T'(Transmission) 'l'(slice legend)
        if 'l' is selected
        '''
        if opts:Vopt=opts
        if isinstance(iSz,int):iSz=[iSz]
        if isinstance(iSz,slice):iSz=list(np.arange(self.ns)[iSz])
        if isinstance(iSz,list):N=len(iSz)
        if isinstance(cmaps,str):cmaps=[cmaps]*3
        cs1,cs2,cs3 = dsp.getCs(cmaps[0],N),dsp.getCs(cmaps[1],N),dsp.getCs(cmaps[2],N)
        plts,legElt=[],{}
        if 'V' in Vopt:
            plts += [[self.x,self.Vz[iSz[i],:].T ,cs1[i]] for i in range(N)]
            legElt[r'$V_z(kV\AA)$']=[cs1[int(N/2)],'-']
        if 'T' in Vopt:
            plts+= [[self.x,self.T.real[iSz[i],:].T,cs2[i]] for i in range(N)]
            plts+= [[self.x,self.T.imag[iSz[i],:].T,cs3[i]] for i in range(N)]
            legElt['$re(T)$']=[cs2[int(N/2)],'-']
            legElt['$im(T)$']=[cs3[int(N/2)],'-']
        return dsp.stddisp(plts,labs=[r'$x(\AA)$',''],#title='Projected potential $V_z$, Transmission function $T$',
            legElt=legElt,**kwargs)

    def Pq_show(self,**kwargs):
        '''Show Propagator function in q space '''
        q  = fft.fftshift(self.q)
        Pq = fft.fftshift(self.Pq)
        plts = [[q,Pq.real,'b','$re$'],
                [q,Pq.imag,'r','$im$']]
        return dsp.stddisp(plts,labs=[r'$q(\AA^{-1})$','$P_q$'],
            title='Propagator',
            **kwargs)

    def Xz_show(self,iZs=1,cmap='jet',**kwargs):
        '''Show wave propagation for slices iZs '''
        if isinstance(iZs,int):iZs=slice(0,self.z.size,iZs)
        z  = self.z[iZs]
        Px = self.psi_xz[iZs,:]
        cs = dsp.getCs(cmap,z.size)
        plts = [ [self.x,Px[i,:],cs[i]] for i in range(z.size)]
        return dsp.stddisp(plts,labs=[r'$x(\AA)$',r'$|\Psi(x)|^2$'],
            imOpt='hc',caxis=[z.min(),z.max()],cmap=cmap,
            **kwargs)

    def Qz_show(self,iZs=1,opts='',cmap='jet',**kwargs):
        '''Show wave propagation in q space for slices iZs
        - opts : O(include Origin), n(normlize integral 1), N(Normalize with max), S(Shape normalize) l(show label)
        '''
        # if isinstance(iZs,int):iZs=slice(0,iZs,self.z.size)
        if isinstance(iZs,int):iZs=slice(0,None,iZs)
        # if isinstance(iZs,list):iZs=slice(0,iZs,self.z.size)
        Pqs = np.zeros((self.psi_qz[iZs,:].shape))
        z   = self.z[iZs]
        Pqs = self.psi_qz[iZs,:].copy()
        if 'S' in opts:
            if 'O' not in opts:Pqs[:,0]=0# do not show central beam
            for i in range(z.size) : Pqs[i,:]/=np.sum(Pqs[i,:])
        else:
            if 'O' not in opts:Pqs[:,0]=0# do not show central beam
            if 'N' in opts:
                for i in range(z.size) : Pqs[i,:]/=Pqs[i,:].max()
            elif 'n' in opts:
                for i in range(z.size) : Pqs[i,:]/=Pqs[i,:].sum()
        q   = fft.fftshift(self.q.copy())
        Pqs = [fft.fftshift(Pqs[i,:]) for i in range(z.size)]
        cs  = dsp.getCs(cmap,z.size)
        if 'l' in opts:
            plts = [[q,Pqs[i],cs[i] ,'$%.1fA$' %z[i]] for i in range(z.size)]
            return dsp.stddisp(plts,labs=[r'$q(\AA^{-1})$',r'$|\Psi(q)|^2$'],
                imOpt='c',caxis=[z.min(),z.max()],cmap=cmap,axPos='V',
                **kwargs)
        else:
            plts = [[q,Pqs[i],cs[i]] for i in range(z.size)]
            return dsp.stddisp(plts,labs=[r'$q(\AA^{-1})$',r'$|\Psi(q)|^2$'],
                imOpt='c',caxis=[z.min(),z.max()],cmap=cmap,axPos='V',
                **kwargs)

    def Bz_show(self,iBs='O',tol=1e-3,cmap='jet',sym_opt=False,plts=[],**kwargs):
        '''Show selected beam iBs as function of thickness(see getB)
        - sym_opt : special display for symmetry pairs
        '''
        # print(iBs)
        iBs,Ib = self.getB(iBs,tol,v=1)#;print(iBs)
        h    = ['%d_{%d}' %(i/self.Nx,i%self.Nx) for i in iBs]
        cs   = dsp.getCs(cmap,iBs.size)
        plts = [[self.z,Ib[:,i],[cs[i],'-'],'$%s$' %h[i]] for i,iB in enumerate(iBs)]
        if sym_opt:
            iBs,Ib = self.getB(-iBs,tol,v=1)
            h    = ['%d_{%d}' %(i/self.Nx,i%self.Nx) for i in iBs]
            plts += [[self.z,Ib[:,i],[cs[i],'--'],'$%s$' %h[i]] for i,iB in enumerate(iBs)]
        return dsp.stddisp(plts,labs=[r'$z(\AA)$',r'$I_b$'],**kwargs)

    def Xxz_show(self,iZs=1,iXs=1,**kwargs):
        '''Show 2D wave propagation solution'''
        if isinstance(iZs,int):iZs=slice(0,-1,iZs)
        if isinstance(iXs,int):iZs=slice(0,-1,iXs)
        x,z = np.meshgrid(self.x,self.z)
        im = [x[iZs,:],z[iZs,:],self.psi_xz[iZs,:]]
        return dsp.stddisp(im=im,labs=[r'$x(\AA)$',r'$z(\AA)$'],
        **kwargs)

    def Qxz_show(self,iZs=1,**kwargs):
        '''Show 2D wave propagation solution'''
        iZs = slice(0,-1,iZs)
        q,z = np.meshgrid(self.q,self.z[iZs])
        Pqs = self.psi_qz[iZs,:].copy()
        Pqs[:,0] = 0
        im = [q,z,Pqs]
        return dsp.stddisp(im=im,labs=[r'$q(\AA^{-1})$',r'$z(\AA)$'],
        **kwargs)
    def Ewald_show(self,deg=10,nh=20,nk=10,relrod=0,**kwargs):
        ''' Displays the Ewald sphere and the reciprocal lattice
        - deg : extent of the ewald sphere in degrees
        - nh,nk : number of reciprocal nodes
        '''
        # Ewald circle
        K  = self.k0
        dt = deg*np.pi/180
        t  = 3*np.pi/2+dt*np.linspace(-1,1,1000)
        # reciprocal lattice
        a1,a2 = lat.get_lattice_vec(lat_type='rect',a=self.ax,b=self.bz)
        b1,b2 = lat.reciprocal_lattice_2D(a1,a2)
        h,k = np.meshgrid(np.arange(-nh,nh+1),np.arange(nk))
        X = h*b1[0]+k*b2[0]
        Z = h*b1[1]+k*b2[1]
        #tilt
        rot = lambda t:np.array([[np.cos(t),np.sin(t)],[-np.sin(t),np.cos(t)]])
        theta = self.tilt*np.pi/180
        X,Z  = rot(-theta).dot(np.stack([X.flatten(),Z.flatten()]))
        plts = []
        if relrod:
            H    = Nz*bz
            zeta = np.linspace(0,0.25/bz,100)
            Fz   = lambda i : 1/(1.1*self.ax)*np.sinc(zeta*H)**2+i/self.ax
            plts +=[[Fz(i),zeta,'b--',''] for i in range(nh)]
        plts += [[K*np.cos(t),K*np.sin(t)+K,'r','']]
        scat  = [X,Z,15,'k']
        # ax1 = np.sqrt(1**2+10**2)*ax
        # bz1 = ax1
        return dsp.stddisp(plts,scat=scat,labs=[r'$k_x(\AA^{-1})$',r'$k_z(\AA^{-1})$'],
            **kwargs)#lw=2,xylims=[0,3,0,3])#,xyTicks=[1/ax1,1/bz1])

    ##########################################
    #### get methods
    def getQ(self):return self.q
    def getI(self):return self.psi_qz[-1,:]
    def getB(self,iBs='Oa',tol=0,v=0):
        ''' get beam as function of thickness
        - iBs :
            - str - 'O'(include origin) a(all)
            - list - the list of indices of the beams
        - tol : select only beams max(I)>tol*I_max
        '''
        Ib = self.psi_qz/self.nx**2/self.dq
        if isinstance(iBs,list):np.array(iBs)
        if isinstance(iBs,str):
            N   = int(self.nx/2)
            iHs = fft.fftshift(np.arange(-N,N))
            if not 'a' in iBs:
                Im  = Ib[:,1:].max()    #;print(Im)
                Imax = Ib.max(axis=0)   #;print(Imax)
                iHs = iHs[Imax>Im*tol]
            iBs=iHs['O' not in iBs:]
        if isinstance(iBs,list):iBs=np.array(iBs)
        if v:
            return iBs,Ib[:,iBs]
        else:
            return Ib[:,iBs]

    ##################################################################
    ###### main computations
    ##################################################################
    def _set_transmission_function(self):
        x,z,f = self.pattern
        nx,ns = x.size,self.ns
        Vz = np.zeros((ns,nx))
        iZs = np.arange(0,ns+1)*int(z.size/ns)#;print(iZs)
        for i_s in range(ns):
            s=slice(iZs[i_s],iZs[i_s+1])
            Vz[i_s,:] = np.array([trapz(f[s,i],z[s]) for i in range(nx)])

        T  = np.exp(1J*self.eps*self.sig*Vz)

        Nx = [self.Nx,1][self.TDS]
        self.x  = np.hstack([x + self.ax*i for i in range(Nx)])
        self.Vz = np.hstack([Vz]*Nx)
        self.T  = np.hstack([T]*Nx)             #;print(self.T.shape)
        self.nx = self.x.size                   #;print(self.nx)

    def _set_propagator(self):
        sg,copt = self.sg,self.copt
        self.dx = self.x[1]-self.x[0]
        self.q  = fft.fftfreq(self.nx,self.dx)
        self.dq = self.q[1]-self.q[0]
        if self.tilt: #isinstance(tilt,np.ndarray) or isinstance(tilt,list):
            kx = self.k0*np.sin(self.tilt*np.pi/180)
            self.Pq = np.exp(self.sg*1J*np.pi*self.dz*(self.q+kx)**2/self.k0)
        else:
            self.Pq = np.exp(self.sg*1J*np.pi*self.dz*self.q**2/self.k0)
        self.nq = int(1/3*self.nx) #prevent aliasing

        if copt:self.Pq[self.nq:-self.nq] = 0

    def set_Psi0(self,iTDS=0):
        Psi  = np.ones(self.x.shape,dtype=complex)
        self.Psi_x = Psi/np.sqrt(np.sum(np.abs(Psi)**2)*self.dx)
        if self.TDS :
            if not iTDS :
                self.psi_xz = np.zeros((0,self.nx))
                self.psi_qz = np.zeros((0,self.nx),dtype=complex)
        else:
            self.psi_xz = np.zeros((0,self.nx))
            self.psi_qz = np.zeros((0,self.nx))
        self.z  = np.array([])
        self.iz = 0

    def propagate(self,nz,iZs=1,iZv=1,opts='q',iTDS=0,v=1):
        '''Propgate over nz slices and save every iZs slices'''
        self.nz+=nz
        nzq,z0 = int(nz/iZs),0
        if self.z.size : z0=self.z.max()
        self.z  = np.hstack([self.z,z0+self.dz+np.arange(nzq)*self.dz*iZs ])
        if 'x' in opts and not iTDS:self.psi_xz = np.vstack([self.psi_xz,np.zeros((nzq,self.nx))])
        if 'q' in opts and not iTDS:self.psi_qz = np.vstack([self.psi_qz,np.zeros((nzq,self.nx))])
        # self.T=fft.fftshfft(self.T)
        for i in range(nz):
            i_s=i%self.ns
            #print(self.T[i_s,:].shape,self.Psi_x.shape)
            self.Psi_q = fft.fft(self.T[i_s,:]*self.Psi_x) #periodic assumption
            if self.copt:self.Psi_q[self.nq:-self.nq] = 0  #prevent aliasing
            self.Psi_x = fft.ifft(self.Pq*self.Psi_q)
            # self.Psi_x = fft.fftshift(fft.ifft(self.Pq*self.Psi_q))
            #save and print out
            msg=''
            if v and (not i%iZv or i==nz-1):
                Ix2 = np.sum(np.abs(self.Psi_x)**2)*self.dx
                Iq2 = np.sum(np.abs(self.Psi_q/self.nx)**2)/self.dq #parseval's theorem of the DFT
                msg+='i=%-4d,islice=%-2d I=%.4f, Iq=%.4f ' %(i,i_s,Ix2,Iq2)
            if not i%iZs :
                if msg and v: msg+='iz=%d, z=%.1f A' %(self.iz, self.z[self.iz])
                if 'x' in opts:
                    self.psi_xz[self.iz,:] = np.abs(self.Psi_x)**2
                if 'q' in opts:
                    if self.TDS:
                        self.psi_qz[self.iz,:] += self.Psi_q
                    else:
                        self.psi_qz[self.iz,:] = np.abs(self.Psi_q)**2
                    self.iz+=1
            if msg:print(colors.green+msg+colors.black)


def plot_v(x,z,fv,ax,bz,xa,za):
    print('..plotting fv...')
    dr=bz/2
    idx = np.abs(x[0]-ax/2)<dr
    xylims = np.hstack([za+dr*np.array([-1,1]),xa+dr*np.array([-1,1])])
    dsp.stddisp(im=[z[:,idx],x[:,idx],fv[:,idx]],
        labs=['$z$','$x$'],imOpt='c',axPos='V',xylims=xylims,opt='p')


##################################################################
###### Base test
##################################################################
if __name__=='__main__':
    print('run tests/,multi_2D/base_test.py to see an example')

    # import wallpp.plane_group as pg
    # import importlib as imp
    # imp.reload(pg)
    # ndeg = 2**8
    # pptype,a,b,angle = 'p1',20,6,90
    # pattern = np.array([[10,2,3],[5,4,3]])
    #
    # #get the potential from wallpaper library
    # p1      = pg.Wallpaper(pptype,a,b,angle,pattern,ndeg=ndeg)
    # pattern = p1.get_potential_grid()
    # mp1 = multi2D(pattern,a,b,keV=200,
    #         Nx=1,dz=b/2,nz=2,ppopt='',#XQZTP
    #         iZs=1,iZv=1)
    # mp1.Tz_show(slice(0,None,1))



import importlib as imp
import pandas as pd, numpy as np
import os,matplotlib,cbf,tifffile,re,glob,easygui
from subprocess import check_output
from matplotlib import rc
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
# from scipy.spatial import ConvexHull, convex_hull_plot_2d
from utils import glob_colors as colors,handler3D as h3d
from utils import physicsConstants as cst
from utils import displayStandards as dsp   #; imp.reload(dsp)
# from scattering.structure_factor import structure_factor3D
# from . import rotating_crystal as rcc       #; imp.reload(rcc)
# from . import postprocess as pp             #; imp.reload(pp)
from . import pets as pt                      #;imp.reload(pt)


class Base_Viewer:
    """
    Viewer class tiff/cbf files (similar to adxv)

    Parameters
    ----------
    figpath
        folder path containing the images
    frame
        frame number to start with
    i
        image index to start with
    cutoff
        brightness cut off
    pargs
        arguments to be passed to :meth:~Base_Viewer.show_im
    thick
        thickness info
    h
        show help


    .. NOTE::
        Works also for a single image
    """
    def __init__(self,figpath:str='',
        frame:Optional[int]=None,i:Optional[int]=0,
        cutoff:int=50,pargs:dict={},
        h:bool=True,
        thick:float=5,
    ):
        self.figpath = figpath
        #### format
        d_fmt = {'cbf':self.load_cbf,'tiff':self.load_tif,'tif':self.load_tif}
        self.supported_fmts = d_fmt.keys()
        self.fmt  = self.find_format(v=h)
        self.load = d_fmt[self.fmt]

        self.figs  = np.sort(glob.glob(self.figpath+'/*.%s' %self.fmt))#;print(self.figs)
        self.nfigs = self._get_nfigs()

        if frame:
            i = min(max(frame,0),self.nfigs)-1
        else:
            frame = i+1
        self.frame  = frame       #starting frame
        self.i      = i           #starting image
        self.inc    = 1           #increment(use 'p' or 'm' to change)
        self.cutoff = cutoff
        self.thick  = thick
        self.fieldNames  = ['thick','frame','cutoff','inc']
        self.mode   = 1
        self.pargs  = pargs
        rc('text', usetex=False)

        self.fig,self.ax = dsp.stddisp(opt='')
        cid = self.fig.canvas.mpl_connect('key_press_event', self)
        self.show()

        self.msg = '''
        'up or right'  : show frame+1
        'down or left' : show frame-1
        ##
        'p' : increase increment rate
        'm' : decrease increment rate
        ##
        'pageup'   : increae cutoff brightness
        'pagedown' : decrease cutoff brightness
        'r'        : reset cutoff brightness
        ##
        'ctrl+t' : increase thickness
        'ctrl+T' : decrease thickness
        ##
        'enter' : change settings
        'S' : save image
        'h' : show help
        '''
        if not 'msgD' in self.__dict__:
            self.msgD=''
        if h:self.show_help()

    def __call__(self, event):
        # print(event.key)
        if event.key in ['up','right']:
            self.i=min(self.i+self.inc,self.nfigs-1)
            self.mode=1
        elif event.key in ['left','down']:
            self.i=max(0,self.i-self.inc)
            self.mode=-1
        self.frame=self.i+1

        if event.key=='S':
            dsp.saveFig(self.get_figname(),ax=self.ax)

        #increment rate
        if event.key=='p':
            self.inc=min(self.inc+1,100)        ;print('increment rate : %d' %self.inc)
        elif event.key=='m':
            self.inc=max(1,self.inc-1)          ;print('increment rate : %d' %self.inc)
        elif event.key=='ctrl+r':
            self.inc=1                          ;print('increment rate : %d' %self.inc)

        #brightness
        dc=25
        if event.key=='pageup':
            self.cutoff=min(self.cutoff+dc,10000)  ;print('cutoff : %d' %self.cutoff)
        elif event.key=='pagedown':
            self.cutoff=max(1,self.cutoff-dc)    ;print('cutoff : %d' %self.cutoff)
        elif event.key=='r':
            self.cutoff=50                      ;print('cutoff : %d' %self.cutoff)

        #thickness
        if event.key=='ctrl+t':
            self.thick+=5                       ;print('thickness : %d' %self.thick)
        elif event.key=='ctrl+T':
            self.thick=max(self.thick-5,5)      ;print('thickness : %d' %self.thick)

        if event.key=='h':self.show_help()
        elif event.key=='enter':self.settings()
        keys = self.call(event)

        update_keys = keys+['enter','ctrl+t','ctrl+T','pageup','pagedown','r','left','right','down','up']
        if event.key in update_keys:
            # self.update_im()
            self.show()

    def settings(self):
        fieldValues = ['%d' %self.__dict__[f] for f in self.fieldNames]
        fieldNames = self.fieldNames.copy()
        self.get_fieldValues(fieldNames,fieldValues)
        dict_fv = multenterbox("Change settings","settings", fieldValues,fieldNames)
        if dict_fv:
            for f in self.fieldNames:
                self.__dict__[f] = int(dict_fv[f])
            self.set_fieldValues(dict_fv)
        self.i = self.frame-1

    def show(self):
        im=self.load(self.figs[self.i])
        self.ax.cla()
        self.show_im(im,**self.pargs)
        self.fig.canvas.draw()
        # self.im = self.ax.get_images()[0]

    def update_im(self):
        fig = self.figs[self.i]
        im = self.load(fig)
        # self.im.set_data(im)
        self.ax.cla()
        dsp.stddisp(im=[im],ax=self.ax,title='image %d' %self.i,
            cmap='gray',caxis=[0,self.cutoff],pOpt='t',
            name=self.get_figname(),**self.pargs)
        # print(self.cutoff)
        self.fig.canvas.draw()

        # figname=os.path.basename(fig)
        print(colors.yellow+fig+colors.black)

    def show_help(self):
        print(colors.green+'shortcuts : '+colors.blue+self.msg+colors.magenta+self.msgD+colors.black)
    def _get_nfigs(self):
        return self.figs.size
    def get_figname(self):
        return self.figs[self.i][:-3]+'.png'

    ###################################
    ##### virtual functions
    ###################################
    def show_im(self,im,**kwargs):
        """The function used to display the frames"""
        print(self.cutoff)
        dsp.stddisp(im=[im],ax=self.ax,title='image %d' %self.i,
            cmap='gray',caxis=[0,self.cutoff],pOpt='tX',
            # xylims=[0,516,0,516],
            name=self.get_figname(),**kwargs)

    def get_fieldValues(self,fieldNames,fieldValues):return None
    def set_fieldValues(self,dict_fv):return None
    def call(self,event):return []

    ###############################################################
    #### misc
    ###############################################################
    def find_format(self,v=1):
        fmts = np.unique([f.split('.')[-1] for f in os.listdir(self.figpath)])
        fmts = [fmt for fmt in fmts if fmt in self.supported_fmts]
        if not len(fmts):
            raise Exception('no supported format found in %s. Supported formats :' %(self.figpath),self.supported_fmts)
        fmt = fmts[0]
        if len(fmts)>1:
            print('warning multiple formats found',fmts)
            print('using %s' %fmt)
        if v:print('%s format detected' %fmt)
        return fmt

    def load_cbf(self,fig):
        try:
            content = cbf.read(fig)
        except:#UnicodeDecodeError
            self.i=self.i+self.mode
            print(colors.red+'error reading file'+colors.black)
            self.import_exp()
            return
        numpy_array_with_data = content.data
        header_metadata = content.metadata
        # print(colors.blue,header_metadata,colors.black)
        return numpy_array_with_data

    def load_tif(self,fig):
        return tifffile.imread(fig)





class Pets_Viewer(Base_Viewer):
    """Pets enhanced viewer

    Parameters
    ----------
    pets
        full path to pets folder
    init
        initial config to display : c(center), r(reflections), p(spot prediction), b(boxes)
    """
    def __init__(self,pets:str=None,
        Smax=0.025,Nmax=13,rot=203,sim=False,
        init:str='',**kwargs,
    ):
        if isinstance(pets,str):pets = pt.Pets(pets)
        if isinstance(pets,pt.Pets):
            if sim:
                figpath = os.path.join(pets.path,'tiff/simulation')
            else:
                figpath = os.path.join(pets.path,'tiff')

        self.pets = pets
        if self.pets:
            vals  = ['center','refl','pred','boxes']
            keys  = ['c','r','p','b']
            alias = ['1','2','3','4']
            self.show_d = dict(zip(keys,vals))
            self.alias  = dict(zip(alias,keys))
            self.show_opt = dict(zip(vals,[c in init for c in keys ]))
            self.Smax,self.Nmax,self.rot=Smax,Nmax,rot
        self.msgD='''
        #### pets info
        '1 or c' : show beam center
        '2 or r' : show detected reflections (in rpl)
        '3 or b' : show boxes
        '4 or p' : show lattice spot locations from info
        '''
        super().__init__(figpath=figpath,**kwargs)

    def call(self,event):
        # k = event.key
        if self.pets:
            if event.key in self.alias.keys():event.key = self.alias[event.key]
            if event.key in self.show_d.keys():
                key = self.show_d[event.key]
                self.show_opt[key] = not self.show_opt[key]

            return list(self.show_d.keys())
        else:
            return []

    def show_im(self,im,**kwargs):
        tle = "frame %d, thickness=%d $\AA$" %(self.frame,self.thick)
        print('ok')
        plts,scat,pp = [],[],[]
        sargs = {'facecolor':'none','edgecolor':(0.7,0.7,0.15),'s':50,'marker':'o'}
        if self.pets:
            # df = self.rpl
            df = self.pets.rpl
            txts = []
            frame = self.i+1
            if self.show_opt['refl']:
                rpl = df.loc[df.F==frame]
                plts+=[[rpl.rpx-0.5,rpl.rpy-0.5,'r+','']]
                txts+=[[x,y,t,'r'] for x,y,t in rpl[['rpx','rpy','hkl']].values]
            if self.show_opt['center']  :
                cen = self.pets.cen.iloc[self.i]
                plts += [[ cen.px-0.5,cen.py-0.5,'b+','']]
            # if self.show_opt['pred'] or self.show_opt['boxes']:
            #     px,py,I,hkl = self.pets.get_kin(frame,
            #         rot=self.rot,thick=self.thick,Smax=self.Smax,Nmax=self.Nmax,pixel=True)
            # if self.show_opt['pred']:
            #     scat  = [px,py]
            # if self.show_opt['boxes']:
            #     npx = 15
            #     pp = [dsp.Rectangle((px0-npx/2,py0-npx/2),npx,npx,facecolor='none',edgecolor='r') for px0,py0 in zip(px,py)]
        print('cutoff',self.cutoff)
        dsp.stddisp(plts,ax=self.ax,patches=pp,scat=scat,im=[im],ms=20,sargs=sargs,
            xylims=[0,516,516,0],
            # xylims=[0,516,0,516],
            texts=txts,
            cmap='gray',caxis=[0,self.cutoff],title=tle,pOpt='tX',**kwargs)






# class Image_viewer:
#     '''a copy of dials.image_viewer'''
#     def __init__(self,file,sym=1,pad=None):
#         basename      = file.split('_')
#         self.im       = int(basename[-1].replace('.cbf',''))
#         self.basename = '_'.join(basename[:-1])
#         file_ids      = [int(f.split('_')[-1].replace('.cbf','')) for f in glob.glob(self.basename+'*.cbf')]
#         if pad :
#             self.pad=pad
#         else:
#             self.pad=int(np.log10(max(file_ids)))+1       #;print(self.pad)
#
#         ls = np.array(check_output("head -n25 %s" %file,shell=True).decode().split('\n'))
#         px_str      = ls[['Pixel_size' in l for l in ls]][0]
#         D_str       = ls[['Detector_distance' in l for l in ls]][0] #;print(D_str)
#         lam_str     = ls[["Wavelength" in l for l in ls]][0]
#
#         self.pxy    = np.array(re.findall("\d+e-\d+", px_str),dtype=float)
#         self.D      = 1 #float(re.findall("\d+.\d+",D_str)[0])
#         self.lam    = float(re.findall("\d+.\d+",lam_str)[0])
#
#         shape = np.array(cbf.read(file).data.shape)
#         self.image = cbf.read(file).data
#
#         if sym:
#             Nx,Ny = np.array(shape/2,dtype=int)
#             self.Nx,self.Ny = Nx,Ny
#             self.pX,self.pY = np.meshgrid(np.arange(-Nx,Nx+1),np.arange(-Ny,Ny+1))
#         else:
#             Nx,Ny = shape
#             # self.pX,self.pY = np.meshgrid(np.arange(Nx),np.arange(Ny))
#             self.pX,self.pY = np.meshgrid(np.arange(Nx),np.flipud(np.arange(Ny)))
#         self.qx,self.qy = self.px2s(self.pX,self.pY)
#
#     def s2px(self,xh):
#         '''xh : recorded coordinates on image'''
#         px,py = self.pxy
#         pX = self.lam*self.D/px*xh[:,0] + self.Nx
#         pY = -self.lam*self.D/py*xh[:,1] + self.Ny
#         return pX,pY
#
#     def px2s(self,pX,pY):
#         px,py = self.pxy
#         qx,qy = px/self.D/self.lam*pX, py/self.D/self.lam*pY
#         return qx,qy
#
#     def show_image(self,im=None,lab='q',stack=1,**kwargs):
#         rc('text', usetex=False)
#         if not im:im = self.im
#         file = "%s_%s.cbf" %(self.basename,str(im).zfill(self.pad))   #;print(filename)
#         # with open(file,'wb') as f:print(file)
#         content = cbf.read(file)
#         image = content.data
#         if stack>1:
#             for i in range(1,stack):
#                 filename = "%s_%s.cbf" %(self.basename,str(im+i).zfill(self.pad))   #;print(filename)
#                 image+=cbf.read(filename).data
#             if 'caxis' in list(kwargs.keys()):
#                 kwargs['caxis'] = list(np.array(kwargs['caxis'])*stack) #; print(kwargs['caxis'])
#         if 'q' in lab:
#             labs = ['$q_x(A^{-1})$','$q_y(A^{-1})$']
#             X,Y = self.qx,-self.qy
#         elif 'p' in lab:
#             labs = ['','']#['$p_x$','$p_y$']
#             X,Y = self.pX,self.pY
#         elif 'x' in lab:
#             labs = ['$x(mm)$','$y(mm)$']
#             px,py = self.pxy
#             X,Y = self.pX*px*1e3,self.pY*py*1e3
#         return dsp.stddisp(im=[X,Y,image],labs=labs,
#             pOpt='ptX',title="%s" %os.path.basename(file),**kwargs)





# class Frames_Viewer(Base_Viewer):
#     '''Viewer for pets'''
#     def __init__(self,pets,thick,Imag,kargs,**sargs):
#         self.pets   = pets
#         self.kargs  = kargs
#         super().__init__(thick=thick,cutoff=Imag,**sargs)
#
#     def get_fieldValues(self,fieldNames,fieldValues):
#         fieldNames+=list(self.kargs.keys())
#         fieldValues+=[str(f) for f in self.kargs.values()]
#     def set_fieldValues(self,dict_fv):
#         float_keys = np.setdiff1d(list(self.kargs.keys()),'opts')
#         for f in float_keys:
#             self.kargs[f] = eval(dict_fv[f])
#         self.kargs['opts']=dict_fv['opts']
#     def _get_nfigs(self):
#         return self.pets.nFrames
#     def get_im(self,**kwargs):
#         self.pets.show_frame(frame=self.i+1,thick=self.thick,Imag=self.cutoff,
#             **self.kargs,**kwargs)
#
#     def call(self,event):
#         chars = 'KkPhr'
#         keys = ['ctrl+K','ctrl+k', 'ctrl+H','ctrl+h','ctrl+g']
#
#         vals = [c in self.kargs['opts'] for c in chars]
#         for i,(c,k) in enumerate(zip(chars,keys)):
#             if event.key==k:
#                 vals[i] = not vals[i]
#
#         self.kargs['opts']='q'+''.join([c for c,val in zip(chars,vals) if val])
#
#         return keys




def multenterbox(msg,title,fieldValues,fieldNames):
    fieldValues = easygui.multenterbox(msg, title, fieldNames,fieldValues)
    while True:
        if fieldValues is None:
            break
        errs = list()
        for n, v in zip(fieldNames, fieldValues):
            if v.strip() == "":errs.append('"{}" is a required field.'.format(n))
        if not len(errs):
            break
        fieldValues = easygui.multenterbox("\n".join(errs), title, fieldNames, fieldValues)
    return dict(zip(fieldNames,fieldValues))
